#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import json
from typing import Any, Dict, List, Union, Optional, Callable

import hjson

from kgforge.core import Resource
from kgforge.core.commons.attributes import sort_attrs
from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import as_jsonld
from kgforge.core.resource import encode


def as_json(data: Union[Resource, List[Resource]], expanded: bool, store_metadata: bool,
            model_context: Optional[Context], metadata_context: Optional[Context],
            context_resolver: Optional[Callable]) -> Union[Dict, List[Dict]]:
    if expanded:
        return as_jsonld(data, "expanded", store_metadata, model_context=model_context,
                         metadata_context=metadata_context, context_resolver=context_resolver)
    else:
        if isinstance(data, Resource):
            return _as_json(data, store_metadata)
        else:
            return [_as_json(x, store_metadata) for x in data]


def from_json(data: Union[Dict, List[Dict]], na: Union[Any, List[Any]]
              ) -> Union[Resource, List[Resource]]:
    nas = na if isinstance(na, List) else [na]
    return _from_json(data, nas)


def _as_json(resource: Resource, store_metadata: bool) -> Dict:
    data = json.loads(hjson.dumpsJSON(resource, default=encode, item_sort_key=sort_attrs))
    _remove_context(data)
    if store_metadata is True and resource._store_metadata:
        data.update(json.loads(hjson.dumpsJSON(resource._store_metadata, item_sort_key=sort_attrs)))
    return data


def _from_json(data: Union[Any, List[Any]], na: List[Any]) -> Any:
    if isinstance(data, List):
        return [_from_json(x, na) for x in data]
    elif isinstance(data, Dict):
        properties = {k: _from_json(v, na) for k, v in data.items() if v not in na}
        return Resource(**properties)
    else:
        return data


def _remove_context(dictionary: dict):
    if isinstance(dictionary, dict):
        for i in list(dictionary):
            if i == "context":
                dictionary.pop("context")
            else:
                if isinstance(dictionary[i], dict):
                    _remove_context(dictionary[i])
                elif isinstance(dictionary[i], list):
                    for x in dictionary[i]:
                        _remove_context(x)




#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from copy import deepcopy


ORGANIZATION = {
    "id": "",
    "type": "Organization",
    "name": "",
    "parentOrganization": {
        "id": "",
        "type": "Organization"
    }
}

PERSON_TEMPLATE = {
    "id": "",
    "type": "Person",
    "address":
        {
            "type": "PostalAddress",
            "postalCode": ["",0],
            "streetAddress": "",
        },
    "birthDate": "9999-12-31",
    "deathDate": "9999-12-31",
    "gender": ["female", "male"],
    "givenName": "",
    "familyName":""
}

EMPLOYEE_TEMPLATE = deepcopy(PERSON_TEMPLATE)
EMPLOYEE_TEMPLATE["type"] = "Employee"
EMPLOYEE_TEMPLATE.update({
    "colleague": {
        "type": "Person",
        "id": "",
        "address":
        {
            "type": "PostalAddress",
            "postalCode": ["",0],
            "streetAddress": "",
        },
        "birthDate": "9999-12-31",
        "deathDate": "9999-12-31",
        "gender": ["female", "male"],
        "givenName": "",
        "familyName":""
    },
    "contractor": {
        "id": "",
        "type": "Organization",
        "name": "",
        "parentOrganization": {
            "id": "",
            "type": "Organization"
        }
    },
    "department": {
        "id": "",
        "type": "Organization",
        "name": "",
        "parentOrganization": {
            "id": "",
            "type": "Organization"
        }
    },
    "startDate": "9999-12-31",
    "worksFor": {
        "id": "",
        "type": ["Organization", "Person"]
    }
})
employee_keys_order = ["id", "type", "address", "birthDate", "colleague", "contractor", "deathDate",
                       "department", "gender", "givenName", "familyName", "startDate", "worksFor"]
EMPLOYEE_TEMPLATE = {k: EMPLOYEE_TEMPLATE[k] for k in employee_keys_order}

ENTITY =  {
    "id": "",
    "type": "Entity"
}

ACTIVITY_TEMPLATE = {
    "id": "",
    "type": "Activity",
    "citation": {
        "id": ""
    },
    "endedAtTime": "9999-12-31T00:00:00",
    "generated": ENTITY,
    "startedAtTime": "9999-12-31T00:00:00",
    "status": "completed",
    "used": ENTITY,
    "validated": False,
    "author": {
        "id": "",
        "type": ["Organization", "Person"]
    }
}

ACTIVITY_TEMPLATE_MANDATORY = {k: v for k, v in ACTIVITY_TEMPLATE.items() if k in ["id", "type", "generated", "status"]}

BUILDING_TEMPLATE = {
    "id": "",
    "type": "Building",
    "description": "",
    "geo": {
        "latitude": 0.0,
        "longitude": 0.0
    },
    "image": {
        "id": ""
    },
    "name": ""
}
BUILDING_TEMPLATE_MANDATORY = {k: v for k, v in BUILDING_TEMPLATE.items() if k in ["id", "type", "description", "name"]}

TYPES_SCHEMAS_MAP = {
    "Activity": "http://www.example.com/ActivityShape",
    "Association": "http://www.example.com/AssociationShape",
    "Building": "http://www.example.com/BuildingShape",
    "Employee": "http://www.example.com/EmployeeShape",
    "Organization": "http://www.example.com/OrganizationShape",
    "Person": "http://www.example.com/PersonShape",
    "PostalAddress": "http://schema.org/PostalAddress",
}


#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from pathlib import Path
from typing import List, Dict, Any, Optional, Callable

from kgforge.core.archetypes import Resolver
from kgforge.core.commons.execution import not_supported
from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping
from kgforge.specializations.resolvers.store_service import StoreService


class OntologyResolver(Resolver):

    def __init__(self, source: str, targets: List[Dict[str, str]], result_resource_mapping: str,
                 **source_config) -> None:
        super().__init__(source,  targets, result_resource_mapping, **source_config)

    @property
    def mapping(self) -> Callable:
        return DictionaryMapping

    @property
    def mapper(self) -> Callable:
        return DictionaryMapper

    def _resolve(self, text: str, target: Optional[str], type: Optional[str],
                 strategy: ResolvingStrategy, limit: Optional[str]) -> Optional[List[Any]]:

        first_filters = f"?id <{self.service.deprecated_property}> \"false\"^^xsd:boolean"
        if type:
            first_filters = f"{first_filters} ; a <{type}>"

        if strategy == strategy.EXACT_MATCH:
            label_filter = f" FILTER (?label = \"{text}\")"
            notation_filter = f" FILTER (?notation = \"{text}\")"
            limit = 1
        else:
            label_filter = f" FILTER regex(?label, \"{text}\", \"i\")"
            notation_filter = f" FILTER regex(?notation, \"{text}\", \"i\")"
            if strategy == strategy.BEST_MATCH:
                limit = 1

        query = """
            CONSTRUCT {{
               ?id a ?type ;
                  label ?label ;
                  prefLabel ?prefLabel ;
                  subClassOf ?subClassOf ;
                  isDefinedBy ?isDefinedBy ;
                  notation ?notation
            }} WHERE {{
              ?id a ?type ;
                  label ?label ; 
              OPTIONAL {{
                ?id subClassOf ?subClassOf ;
                  prefLabel ?prefLabel ;  
                  isDefinedBy ?isDefinedBy ;
                  notation ?notation
              }}
              {{
                SELECT * WHERE {{
                  {{ {0} ; label ?label {1} }} UNION
                  {{ {0} ; notation ?notation {2} }}
                }} LIMIT {3}
              }}
            }}
            """.format(first_filters, label_filter, notation_filter, limit)

        expected_fields = ["type", "label", "prefLabel", "subClassOf", "isDefinedBy", "notation"]

        return self.service.perform_query(query, target, expected_fields, limit)

    @staticmethod
    def _service_from_directory(dirpath: Path, targets: Dict[str, str]) -> Any:
        not_supported()

    @staticmethod
    def _service_from_store(store: Callable, targets: Dict[str, str], **store_config) -> StoreService:
        return StoreService(store, targets, **store_config)



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import json
import re
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

from kgforge.core import Resource
from kgforge.core.archetypes import Mapping, Model
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import ValidationError


class DemoModel(Model):
    """An example to show how to implement a Model and to demonstrate how it is used."""

    def __init__(self, source: str, **source_config) -> None:
        super().__init__(source, **source_config)

    # Vocabulary.

    def _prefixes(self) -> Dict[str, str]:
        return self.service.namespaces

    def _types(self) -> List[str]:
        return [x["label"] for x in self.service.vocabulary["Class"]]

    def context(self) -> Context:
        ctx = {x["label"]: x["id"] for k, v in self.service.vocabulary.items() for x in v}
        return Context({"@context": ctx})

    # Templates.

    def _template(self, type: str, only_required: bool) -> Dict:
        # TODO DKE-148.
        print("<info> DemoModel does not distinguish values and constraints in templates for now.")
        # TODO DKE-148.
        print("<info> DemoModel does not automatically include nested schemas for now.")
        if only_required:
            # TODO DKE-148.
            print("<info> DemoModel does not support keeping only required properties for now.")
        type_expanded = self.service.expand(type)
        schema = self.service.schema(type_expanded)
        return self.service.compact(schema)

    # Mappings.

    def _sources(self) -> List[str]:
        dirpath = Path(self.source, "mappings")
        return [x.stem for x in dirpath.iterdir() if x.is_dir()]

    def _mappings(self, source: str) -> Dict[str, List[str]]:
        dirpath = Path(self.source, "mappings", source)
        mappings = {}
        if dirpath.is_dir():
            for x in dirpath.glob("*/*.hjson"):
                mappings.setdefault(x.stem, []).append(x.parent.name)
        else:
            raise ValueError("unrecognized source")
        return mappings

    def mapping(self, entity: str, source: str, type: Callable) -> Mapping:
        filename = f"{entity}.hjson"
        filepath = Path(self.source, "mappings", source, type.__name__, filename)
        if filepath.is_file():
            return type.load(filepath)
        else:
            raise ValueError("unrecognized entity type or source")

    # Validation.

    def _validate_one(self, resource: Resource) -> None:
        # If the resource is not typed, AttributeError is raised: run() sets _validated to False.
        type_ = resource.type
        type_expanded = self.service.expand(type_)
        schema = self.service.schema(type_expanded)
        result, reason = self.service.check(resource, schema)
        if reason is not None:
            raise ValidationError(reason)

    # Utils.

    @staticmethod
    def _service_from_directory(dirpath: Path, context_iri: str, **dir_config):
        return ModelLibrary(dirpath)


class ModelLibrary:
    """Simulate a third-party library handling interactions with the data used by the model."""

    def __init__(self, dirpath: Path):
        with (dirpath / "namespaces.json").open() as f:
            self.namespaces: Dict[str, str] = json.load(f)
        with (dirpath / "vocabulary.json").open() as f:
            self.vocabulary: Dict = json.load(f)
        with (dirpath / "schemas.json").open() as f:
            self.schemas: Dict = json.load(f)

    def expand(self, value: str) -> str:
        return next(x["id"] for x in self.vocabulary["Class"] if x["label"] == value)

    def compact(self, value: Union[str, Dict]) -> Union[str, Dict]:
        if isinstance(value, Dict):
            return {self.compact(k): self.compact(v) for k, v in value.items()}
        else:
            return re.sub("[a-z]+:", "", value)

    def schema(self, type_expanded: str) -> Dict:
        return self.schemas[type_expanded]

    def check(self, resource: Resource, schema: Dict) -> Tuple[bool, Optional[str]]:
        result, reason = (True, None)
        for k, v in schema.items():
            compacted = self.compact(k)
            if isinstance(v, Dict):
                nested = getattr(resource, compacted)
                result, reason = self.check(nested, v)
                if result is False:
                    return result, reason
            else:
                rule = f"{v}(resource, '{compacted}')"
                try:
                    checked = eval(rule, {}, {"resource": resource})
                except (TypeError, NameError, SyntaxError):
                    pass
                else:
                    if not checked:
                        return False, f"{compacted} is missing"
        else:
            return result, reason



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from typing import Callable, Dict, List, Optional

from rdflib.plugins.sparql.parser import Query

from kgforge.core.archetypes import Store
from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.core.conversions.json import as_json


class StoreService:

    def __init__(self, store: Callable, targets: Dict[str, str], **store_config):
        self.sources: Dict[str, Store] = dict()
        for target, bucket in targets.items():
            store_config.update(bucket=bucket)
            self.sources[target] = store(**store_config)
        self.deprecated_property = "https://bluebrain.github.io/nexus/vocabulary/deprecated"

    def perform_query(self, query: str, target: str, expected_fields: List[str],
                      limit: int) -> Optional[List[Dict]]:
        # FIXME: there is no way to provide a ranking among the results in the different targets
        """Executes the query in the different sources.

        Args:
            query: str, a CONSTRUCT query
            target: a target identifier, if None, it will runt the query in all available targets
            expected_fields: a list of names of the expected fields
            limit: the maximum number of results to return

        Returns:
            List of results in JSON format
        """
        if target:
            if target in self.sources:
                resources = self.sources[target].sparql(query, debug=False, limit=limit)
            else:
                raise ValueError(f"Target {target} not identified")
        else:
            resources = list()
            for source in self.sources.values():
                resources.extend(source.sparql(query, debug=False, limit=limit))

        if len(resources) > 0:
            return [format_response(r, expected_fields) for r in resources]
        else:
            return None


def format_response(resource, mandatory_fields):
    json_data = as_json(resource, expanded=False, store_metadata=False,
                        model_context=None, metadata_context=None, context_resolver=None)
    for field in mandatory_fields:
        if field not in json_data:
            json_data[field] = None
    return json_data



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import re
from importlib import import_module
from typing import Callable

from kgforge.core.commons.exceptions import ConfigurationError


def import_class(configuration: str, forge_module_name: str) -> Callable:
    # Example use:
    #   - import_class("DemoModel", "models")
    #   - import_class("CustomModel from package.models", "models")
    #   - import_class("CustomModel from package.models.custom_model", "models")
    archetype = forge_module_name[:-1].capitalize()
    matched = re.match("^([A-Za-z]+)(?: from ([a-z_.]+))?$", configuration)
    if matched:
        try:
            forge_module_import = f"kgforge.specializations.{forge_module_name}"
            class_name, module_import = matched.groups(default=forge_module_import)
            module = import_module(module_import)
            return getattr(module, class_name)
        except ModuleNotFoundError:
            raise ConfigurationError(f"{archetype} module not found for '{configuration}'")
        except AttributeError:
            raise ConfigurationError(f"{archetype} class not found for '{configuration}'")
    else:
        raise ConfigurationError(f"incorrect {archetype} configuration for '{configuration}'")



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import hjson
from pandas import DataFrame

from kgforge.core import Resource
from kgforge.core.archetypes import Mapping
from kgforge.core.commons.attributes import repr_class, sort_attrs
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import ConfigurationError, ValidationError
from kgforge.core.commons.execution import not_supported, run
from kgforge.core.commons.imports import import_class


class Model(ABC):

    # See demo_model.py in kgforge/specializations/models/ for a reference implementation.

    # POLICY Methods of archetypes, except __init__, should not have optional arguments.

    # POLICY Implementations should be declared in kgforge/specializations/models/__init__.py.
    # POLICY Implementations should not add methods but private functions in the file.
    # TODO Move from BDD to classical testing to have a more parameterizable test suite. DKE-135.
    # POLICY Implementations should pass tests/specializations/models/demo_model.feature tests.

    def __init__(self, source: str, **source_config) -> None:
        # POLICY Model data access should be lazy, unless it takes less than a second.
        # POLICY There could be data caching but it should be aware of changes made in the source.
        self.source: str = source
        self.service: Any = self._initialize_service(self.source, **source_config)

    def __repr__(self) -> str:
        return repr_class(self)

    # Vocabulary.

    def prefixes(self, pretty: bool) -> Optional[Dict[str, str]]:
        prefixes = sorted(self._prefixes().items())
        if pretty:
            print("Used prefixes:")
            df = DataFrame(prefixes)
            formatters = {x: f"{{:<{df[x].str.len().max()}s}}".format for x in df.columns}
            print(df.to_string(header=False, index=False, formatters=formatters))
        else:
            return dict(prefixes)

    def _prefixes(self) -> Dict[str, str]:
        not_supported()

    def types(self, pretty: bool) -> Optional[List[str]]:
        types = sorted(self._types())
        if pretty:
            print(*["Managed entity types:", *types], sep="\n   - ")
        else:
            return types

    @abstractmethod
    def _types(self) -> List[str]:
        # POLICY Should return managed types in their compacted form (i.e. not IRI nor CURIE).
        pass

    def context(self) -> Context:
        # POLICY Should return the context of the Model.
        pass

    def resolve_context(self, iri: str) -> Dict:
        # POLICY Should retrieve the resolved context as dictionary
        not_supported()

    def _generate_context(self) -> Dict:
        # POLICY Should generate the Context from the Model data.
        not_supported()

    # Templates.

    def template(self, type: str, only_required: bool, output: str) -> Optional[Dict]:
        schema = self._template(type, only_required)
        if output == "hjson":
            print(hjson.dumps(schema, indent=4, item_sort_key=sort_attrs))
        elif output == "json":
            print(hjson.dumpsJSON(schema, indent=4, item_sort_key=sort_attrs))
        elif output == "dict":
            return json.loads(hjson.dumpsJSON(schema, item_sort_key=sort_attrs))
        else:
            raise ValueError("unrecognized output")

    @abstractmethod
    def _template(self, type: str, only_required: bool) -> Dict:
        # POLICY Should raise ValueError if 'type' is not managed by the Model.
        # POLICY Each nested typed resource should have its template included.
        # POLICY The value of 'type' and properties should be compacted (i.e. not IRI nor CURIE).
        pass

    # Mappings.

    def sources(self, pretty: bool) -> Optional[List[str]]:
        sources = sorted(self._sources())
        if pretty:
            print(*["Data sources with managed mappings:", *sources], sep="\n   - ")
        else:
            return sources

    def _sources(self) -> List[str]:
        # The discovery strategy cannot be abstracted as it depends on the Model data organization.
        not_supported()

    def mappings(self, source: str, pretty: bool) -> Optional[Dict[str, List[str]]]:
        mappings = {k: sorted(v) for k, v in
                    sorted(self._mappings(source).items(), key=lambda kv: kv[0])}
        if pretty:
            print("Managed mappings for the data source per entity type and mapping type:")
            for k, v in mappings.items():
                print(*[f"   - {k}:", *v], sep="\n        * ")
        else:
            return mappings

    def _mappings(self, source: str) -> Dict[str, List[str]]:
        # POLICY Should raise ValueError if 'source' is not managed by the Model.
        # POLICY Keys should be managed resource types with mappings for the given data source.
        # POLICY Values should be available mapping types for the resource type.
        # The discovery strategy cannot be abstracted as it depends on the Model data organization.
        not_supported()

    def mapping(self, entity: str, source: str, type: Callable) -> Mapping:
        # POLICY Should raise ValueError if 'entity' or 'source' is not managed by the Model.
        # The selection strategy cannot be abstracted as it depends on the Model data organization.
        not_supported()

    # Validation.

    def schema_id(self, type: str) -> str:
        # POLICY Should retrieve the schema id of the given type.
        not_supported()

    def validate(self, data: Union[Resource, List[Resource]],
                 execute_actions_before: bool) -> None:
        # Replace None by self._validate_many to switch to optimized bulk validation.
        run(self._validate_one, None, data, execute_actions=execute_actions_before,
            exception=ValidationError, monitored_status="_validated")

    def _validate_many(self, resources: List[Resource]) -> None:
        # Bulk validation could be optimized by overriding this method in the specialization.
        # POLICY Should reproduce self._validate_one() and execution._run_one() behaviours.
        not_supported()

    @abstractmethod
    def _validate_one(self, resource: Resource) -> None:
        # POLICY Should notify of failures with exception ValidationError including a message.
        pass

    # Utils.

    def _initialize_service(self, source: str, **source_config) -> Any:
        # Model data could be loaded from a directory, an URL, or a Store.
        # Initialize the access to the model data according to the source type.
        # POLICY Should not use 'self'. This is not a function only for the specialization to work.
        origin = source_config.pop("origin")
        context_config = source_config.pop("context", {})
        context_iri = context_config.get("iri", None)
        if origin == "directory":
            dirpath = Path(source)
            return self._service_from_directory(dirpath, context_iri)
        elif origin == "url":
            return self._service_from_url(source, context_iri)
        elif origin == "store":
            store = import_class(source, "stores")
            return self._service_from_store(store, context_config, **source_config)
        else:
            raise ConfigurationError(f"unrecognized Model origin '{origin}'")

    @staticmethod
    @abstractmethod
    def _service_from_directory(dirpath: Path, context_iri: Optional[str]) -> Any:
        pass

    @staticmethod
    def _service_from_url(url: str, context_iri: Optional[str]) -> Any:
        raise NotImplementedError()

    @staticmethod
    def _service_from_store(store: Callable, context_config: Optional[dict],
                            **source_config) -> Any:
        raise NotImplementedError()



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import Dict, List


def with_defaults(original: Dict, other: Dict, original_key: str, other_key: str,
                  keys: List[str]) -> None:
    """Update 'original' with 'other' 'keys' unless keys value is different in both dictionaries."""

    if original[original_key] == other[other_key]:
        for x in keys:
            if x not in original:
                original[x] = other[x]



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import json
from itertools import chain
from pathlib import Path
from typing import Callable, Dict, Iterator, List, Optional

from kgforge.core.archetypes import Resolver
from kgforge.core.commons.exceptions import ConfigurationError
from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping


class DemoResolver(Resolver):
    """An example to show how to implement a Resolver and to demonstrate how it is used."""

    def __init__(self, source: str, targets: List[Dict[str, str]], result_resource_mapping: str,
                 **source_config) -> None:
        super().__init__(source, targets, result_resource_mapping, **source_config)

    @property
    def mapping(self) -> Callable:
        return DictionaryMapping

    @property
    def mapper(self) -> Callable:
        return DictionaryMapper

    def _resolve(self, text: str, target: Optional[str], type: Optional[str],
                 strategy: ResolvingStrategy, limit: Optional[str]) -> Optional[List[Dict[str, str]]]:
        if target is not None:
            data = self.service[target]
        else:
            data = chain.from_iterable(self.service.values())

        if type is not None:
            data = (x for x in data if x.get("type", None) == type)

        if strategy == ResolvingStrategy.EXACT_MATCH:
            try:
                return next(x for x in data
                            if text == x["label"] or ("acronym" in x and text == x["acronym"]))
            except StopIteration:
                return None
        else:
            results = [(_dist(x["label"], text), x) for x in data
                       if text and text in x["label"] or ("acronym" in x and text in x["acronym"])]
            if results:
                ordered = sorted(results, key=lambda x: x[0])
                if strategy == ResolvingStrategy.BEST_MATCH:
                    return ordered[0][1]
                else:
                    # Case: ResolvingStrategy.ALL_MATCHES.
                    return [x[1] for x in ordered]
            else:
                return None

    @staticmethod
    def _service_from_directory(dirpath: Path, targets: Dict[str, str]
                                ) -> Dict[str, List[Dict[str, str]]]:
        return {target: list(_load(dirpath, filename)) for target, filename in targets.items()}


def _dist(x: str, y: str) -> int:
    return len(x) - len(y)


def _load(dirpath: Path, filename: str) -> Iterator[Dict[str, str]]:
    filepath = dirpath / filename
    if filepath.is_file():
        with filepath.open() as f:
            yield from json.load(f)
    else:
        raise ConfigurationError("<source>/<bucket> should be a valid file path")



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import Any, Dict


class DictWrapper(dict):

    # TODO Should be immutable. DKE-146.

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    # No need to define __repr__ and __str__ as the ones from dict are suitable.


def wrap_dict(data: Dict) -> DictWrapper:
    if isinstance(data, Dict):
        return _wrap(data)
    else:
        raise TypeError("not a dictionary")


def _wrap(data: Any) -> Any:
    if isinstance(data, Dict):
        return DictWrapper({k: _wrap(v) for k, v in data.items()})
    else:
        return data



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
import os
from pathlib import Path
from typing import Dict, Tuple

from pyshacl import Validator
from rdflib import Graph, URIRef
from rdflib.util import guess_format

from kgforge.core.commons.context import Context
from kgforge.specializations.models.rdf.node_properties import NodeProperties
from kgforge.specializations.models.rdf.service import RdfService, ShapesGraphWrapper


class DirectoryService(RdfService):

    def __init__(self, dirpath: Path, context_iri: str) -> None:
        self._graph = load_rdf_files(dirpath)
        self._sg = ShapesGraphWrapper(self._graph)
        super().__init__(self._graph, context_iri)

    def schema_source_id(self, schema_iri: str) -> str:
        # FIXME should return the file path where the schema is in
        return schema_iri

    def materialize(self, iri: URIRef) -> NodeProperties:
        sh = self._sg.lookup_shape_from_node(iri)
        predecessors = set()
        props, attrs = sh.traverse(predecessors)
        if props:
            attrs["properties"] = props
        return NodeProperties(**attrs)

    def _validate(self, iri: str, data_graph: Graph) -> Tuple[bool, Graph, str]:
        validator = Validator(data_graph, shacl_graph=self._graph)
        return validator.run()

    def resolve_context(self, iri: str) -> Dict:
        if iri in self._context_cache:
            return self._context_cache[iri]
        else:
            try:
                context = Context(iri)
            except FileNotFoundError as e:
                raise ValueError(e)
            else:
                self._context_cache.update({iri: context.document})
                return context.document

    def generate_context(self) -> Dict:
        return self._generate_context()

    def _build_shapes_map(self) -> Dict:
        query = """
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX sh: <http://www.w3.org/ns/shacl#>
            SELECT ?type ?shape WHERE {
                { ?shape sh:targetClass ?type .}
                UNION {
                    SELECT (?shape as ?type) ?shape WHERE {
                        ?shape a sh:NodeShape .
                        ?shape a rdfs:Class
                    }
                }
            } ORDER BY ?type"""
        res = self._graph.query(query)
        return {row["type"]: row["shape"] for row in res}


def load_rdf_files(path: Path) -> Graph:
    memory_graph = Graph()
    extensions = [".ttl", ".n3", ".json", ".rdf"]
    for f in path.rglob(os.path.join("*.*")):
        if f.suffix in extensions:
            file_format = guess_format(f.name)
            if file_format is None:
                file_format = "json-ld"
            memory_graph.parse(f.as_posix(), format=file_format)
    return memory_graph



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import os
from typing import Callable, List, Union, Dict, Optional
from uuid import uuid4

import pytest
from pytest_bdd import given, parsers, then, when

from kgforge.core import Resource
from kgforge.core.commons.actions import Action
from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import _merge_jsonld, Form


def do(fun: Callable, data: Union[Resource, List[Resource]], *args) -> None:
    if isinstance(data, List) and all(isinstance(x, Resource) for x in data):
        for x in data:
            fun(x, *args)
    elif isinstance(data, Resource):
        fun(data, *args)
    else:
        raise TypeError("not a Resource nor a list of Resource")


# It seems using 'given' with both 'fixture' and 'target_fixture' does not work.
# This implies that (in)valid resources creation steps need their dedicated function.


# Resource(s) creation.

def resource(valid: bool, index: int = 0) -> Resource:
    rid = str(uuid4())
    r = Resource(type="Person", id=rid)
    if valid:
        r.name = f"resource {index}"
    return r


def resources(valid: bool) -> List[Resource]:
    r1 = resource(valid, 0)
    r2 = resource(valid, 1)
    return [r1, r2]


@given("A valid resource.", target_fixture="data")
def valid_resource():
    return resource(True)


@given("Valid resources.", target_fixture="data")
def valid_resources():
    return resources(True)


@given("An invalid resource.", target_fixture="data")
def invalid_resource():
    return resource(False)


@given("Invalid resources.", target_fixture="data")
def invalid_resources():
    return resources(False)


# Resource(s) modifications.

@when("I modify the resource.")
def modify_resource(data):
    data.name = "other"


# Resource(s) verifications.

@then(parsers.parse("The '{attr}' status of {} resource should be '{value}'."))
def check_resource_status(data, attr, value):
    def fun(x): assert str(getattr(x, attr)) == value
    do(fun, data)


# Action(s) verifications.

def check_report(capsys, rc, err, msg, op):
    out = capsys.readouterr().out[:-1]
    heads = {
        "": "",
        "s": "<count> 2\n",
    }
    head = heads[rc]
    tails = {
        None: f"False\n<error> {msg}",
        " not": "True",
    }
    tail = tails[err]
    assert out == f"{head}<action> {op}\n<succeeded> {tail}"


@then("I should be able to access the report of the action on a resource.")
def check_action(data):
    def fun(x): assert isinstance(x._last_action, Action)
    do(fun, data)


@then(parsers.parse("The report should say that the operation was '{value}'."))
def check_action_operation(data, value):
    def fun(x): assert x._last_action.operation == value
    do(fun, data)


@then(parsers.parse("The report should say that the operation success is '{value}'."))
def check_action_success(data, value):
    def fun(x): assert str(x._last_action.succeeded) == value
    do(fun, data)


@then(parsers.parse("The report should say that the error was '{value}'."))
def check_action_error(data, value):
    def fun(x): assert str(x._last_action.error) == value
    do(fun, data)


# Fixtures for Resource to JSON-LD conversion and vice versa

@pytest.fixture
def custom_context():
    return {
        "@context": {
            "@base": "http://example.org/",
            "Person": "http://xmlns.com/foaf/0.1/Person",
            "name": "http://xmlns.com/foaf/0.1/name",
        }
    }


@pytest.fixture
def metadata_context() -> Context:
    document = {
        "deprecated": "https://store.net/vocabulary/deprecated",
        "version": "https://store.net/vocabulary/version"
    }
    return Context(document, "http://store.org/metadata.json")


@pytest.fixture
def metadata_data_compacted():
    return {
        "deprecated": False,
        "version": 1
    }


@pytest.fixture
def metadata_data_expanded():
    return {
        "https://store.net/vocabulary/deprecated": False,
        "https://store.net/vocabulary/version": 1
    }


@pytest.fixture
def person(custom_context):
    return Resource(context=custom_context, type="Person", name="Jami Booth")


@pytest.fixture
def organization(registered_person_custom_context, store_metadata_value):
    contribution = Resource(type="Organization", name="Reichel Inc",
                            founder=registered_person_custom_context)
    return contribution


@pytest.fixture
def building():
    type_ = "Building"
    name = "The Empire State Building"
    description = "The Empire State Building is a 102-story landmark in New York City."
    image = "http://www.civil.usherbrooke.ca/cours/gci215a/empire-state-building.jpg"
    geo = {"latitude": "40.75"}
    return Resource(type=type_, name=name, description=description, image=image, geo=geo)


@pytest.fixture(scope="session")
def context_file_path():
    return os.sep.join((os.path.abspath("."), "tests/data/shacl-model/context.json"))


@pytest.fixture(scope="session")
def context_iri_file(context_file_path):
    return f"file://{context_file_path}"


@pytest.fixture(scope="session")
def context_iri():
    return f"http://example.org/context"


@pytest.fixture(scope="session")
def model_context(context_iri_file) -> Context:
    return Context(context_iri_file, context_iri_file)


@pytest.fixture(scope="session")
def model_prefixes():
    return {
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "prov": "http://www.w3.org/ns/prov#",
        "schema": "http://schema.org/",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "nsg": "https://neuroshapes.org/",
        "owl": "http://www.w3.org/2002/07/owl#",
        "xsd": "http://www.w3.org/2001/XMLSchema#"
    }

@pytest.fixture
def building_jsonld(metadata_context, metadata_data_compacted, metadata_data_expanded):
    def _make_jsonld_expanded(resource, store_metadata, context):
        data = dict()
        if hasattr(resource, "id"):
            data["@id"] = resource.id
        ctx = Context(resource.context) if hasattr(resource, "context") else Context(context)
        latitude_term = ctx.terms.get("latitude")
        if latitude_term.type:
            latitude_node = {
                "@type": latitude_term.type,
                "@value": resource.geo["latitude"]
            }
        else:
            latitude_node = resource.geo["latitude"]
        geo_expanded = {
            latitude_term.id: latitude_node
        }
        data.update({
            "@type": ctx.expand(resource.type),
            ctx.expand("description"): resource.description,
            ctx.expand("geo"): geo_expanded,
            ctx.expand("image"): {"@id": resource.image},
            ctx.expand("name"): resource.name
        })
        if store_metadata and resource._store_metadata is not None:
            data.update(metadata_data_expanded)
        return data

    def _make_jsonld_compacted(resource, store_metadata, context):
        data = dict()
        data_context = resource.context if hasattr(resource, "context") else context
        if store_metadata and resource._store_metadata is not None:
            metadata_context_output = metadata_context.iri if metadata_context.is_http_iri() else metadata_context.document["@context"]
            data["@context"] = _merge_jsonld(data_context, metadata_context_output)
        else:
            data["@context"] = data_context
        if hasattr(resource, "id"):
            data["@id"] = resource.id
        data.update({
            "@type": resource.type,
            "description": resource.description,
            "geo": resource.geo,
            "image": resource.image,
            "name": resource.name
        })
        if store_metadata and resource._store_metadata is not None:
            data.update(metadata_data_compacted)
        return data

    def _make_jsonld(rsrc: Resource, form: str, store_metadata: bool,
                     context: Optional[Union[Dict, List, str]]):
        if form is Form.COMPACTED.value:
            return _make_jsonld_compacted(rsrc, store_metadata, context)
        elif form is Form.EXPANDED.value:
            return _make_jsonld_expanded(rsrc, store_metadata, context)
    return _make_jsonld



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from .bluebrain_nexus import BlueBrainNexus
from .demo_store import DemoStore



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import List, Union

from kgforge.core import Resource
from kgforge.core.commons.actions import LazyAction
from kgforge.core.commons.execution import catch
from kgforge.core.forge import KnowledgeGraphForge


# POLICY _set() should be only called as the last statement to ensure atomicity in case of errors.


class Dataset(Resource):
    """An opinionated high-level class based on Resource to handle datasets."""

    _RESERVED = {"_forge", "add_parts", "add_distribution", "add_contribution", "add_generation",
                 "add_derivation", "add_invalidation", "add_files",
                 "download"} | Resource._RESERVED

    # No catching of exceptions so that no incomplete instance is created if an error occurs.
    # This is a best practice in Python for __init__().
    def __init__(self, forge: KnowledgeGraphForge, type: str = "Dataset", **properties) -> None:
        super().__init__(**properties)
        self._forge: KnowledgeGraphForge = forge
        self.type: str = type

    @catch
    def add_parts(self, resources: List[Resource], versioned: bool = True) -> None:
        """Make resources part of the dataset."""
        keep = ["id", "type", "name", "distribution.contentUrl"]
        parts = self._forge.reshape(resources, keep, versioned)
        _set(self, "hasPart", parts)

    @catch
    def add_distribution(self, path: str, content_type: str = None) -> None:
        # path: FilePath.
        """Add a downloadable form of the dataset."""
        action = self._forge.attach(path, content_type)
        _set(self, "distribution", action)

    @catch
    def add_contribution(self, agent_id: str, **kwargs) -> None:
        # agent: IRI.
        """Add information on the contribution of an agent during the generation of the dataset."""
        agent = Resource(type="Agent", id=agent_id)
        contribution = Resource(type="Contribution", agent=agent, **kwargs)
        _set(self, "contribution", contribution)

    @catch
    def add_generation(self, **kwargs) -> None:
        """Add information on the activity which has resulted in the generation of the dataset."""
        if not kwargs:
            raise TypeError("at least one argument should be given")
        generation = Resource(type="Generation", **kwargs)
        _set(self, "generation", generation)

    @catch
    def add_derivation(self, resource: Resource, versioned: bool = True, **kwargs) -> None:
        """Add information on the derivation of an entity resulting in the dataset."""
        keep = ["id", "type", "name"]
        entity = self._forge.reshape(resource, keep, versioned)
        derivation = Resource(type="Derivation", entity=entity, **kwargs)
        _set(self, "derivation", derivation)

    @catch
    def add_invalidation(self, **kwargs) -> None:
        """Add information on the invalidation of the dataset."""
        if not kwargs:
            raise TypeError("at least one argument should be given")
        invalidation = Resource(type="Invalidation", **kwargs)
        _set(self, "invalidation", invalidation)

    @catch
    def add_files(self, path: str, content_type: str = None) -> None:
        # path: DirPath.
        """Add (different) files as parts of the dataset."""
        action = self._forge.attach(path, content_type)
        distribution = Resource(distribution=action)
        _set(self, "hasPart", distribution)

    @catch
    def download(self, source: str, path: str, overwrite: bool = False) -> None:
        # path: DirPath.
        """Download the distributions of the dataset or the files part of the dataset."""
        if source == "distributions":
            follow = "distribution.contentUrl"
        elif source == "parts":
            follow = "hasPart.distribution.contentUrl"
        else:
            raise ValueError("unrecognized source")
        self._forge.download(self, follow, path, overwrite)


def _set(dataset: Dataset, attr: str, data: Union[Resource, List[Resource], LazyAction]) -> None:
    if hasattr(dataset, attr):
        value = getattr(dataset, attr)
        if isinstance(value, List):
            if isinstance(data, List):
                value.extend(data)
            else:
                value.append(data)
        else:
            if isinstance(data, List):
                new = [value, *data]
            else:
                new = [value, data]
            setattr(dataset, attr, new)
    else:
        setattr(dataset, attr, data)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import pytest
import os
from uuid import uuid4
from urllib.parse import urljoin
from urllib.request import pathname2url

from kgforge.core import KnowledgeGraphForge, Resource
from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import _merge_jsonld
from kgforge.core.wrappings.dict import wrap_dict


@pytest.fixture
def forge():
    config = {
        "Model": {
            "name": "DemoModel",
            "origin": "directory",
            "source": "tests/data/demo-model/",
        },
        "Store": {
            "name": "DemoStore",
        },
    }
    return KnowledgeGraphForge(config)


@pytest.fixture
def r1():
    return Resource(id="123", type="Type", p1="v1a", p2="v2a")


@pytest.fixture
def r2():
    return Resource(id="345", type="Type", p1="v1b", p2="v2b")


@pytest.fixture
def r3(r1):
    return Resource(id="678", type="Other", p3="v3c", p4=r1)


@pytest.fixture
def r4(r2):
    return Resource(id="912", type="Other", p3="v3d", p4=r2)


@pytest.fixture
def r5():
    return Resource(p5="v5e", p6="v6e")


# Fixtures for Resource to JSON-LD conversion and vice versa


@pytest.fixture
def store_metadata_context():
    return {
        "meta": "https://store.net/vocabulary/",
        "deprecated": "meta:deprecated",
        "version": "meta:version"
    }


@pytest.fixture
def store_metadata_value(store_metadata_context, metadata_data_compacted):
    data = {"context": store_metadata_context}
    data.update(metadata_data_compacted)
    return data


@pytest.fixture
def make_registered():
    def _make_registered(r: Resource, metadata, base=None):
        if base:
            r.id = f"{urljoin('file:', pathname2url(os.getcwd()))}/{str(uuid4())}"
        else:
            r.id = str(uuid4())
        if metadata:
            metadata["id"] = r.id
            r._store_metadata = wrap_dict(metadata)
        return r
    return _make_registered


@pytest.fixture
def registered_person_custom_context(person, custom_context, store_metadata_value):
    person.id = str(uuid4())
    store_metadata_value["id"] = person.id
    person._store_metadata = wrap_dict(store_metadata_value)
    return person


@pytest.fixture
def organization_jsonld_compacted(custom_context, model_context, metadata_data_compacted):
    def _make_jsonld_compacted(r, store_metadata):
        data = dict()
        data["@context"] = model_context.iri if model_context.is_http_iri() else model_context.document["@context"]
        if hasattr(r, "id"):
            data["@id"] = r.id
        nested_context = Context(custom_context)
        data.update({
            "@type": r.type,
            "name": r.name,
            "founder": {
                "@type": r.founder.type,
                "@id": nested_context.resolve(r.founder.id),
                "name": r.founder.name
            }
        })
        if hasattr(r, "id") and store_metadata:
            data.update(metadata_data_compacted)
        return data
    return _make_jsonld_compacted


@pytest.fixture
def person_jsonld_compacted():
    def _make_jsonld_compacted(resource, store_metadata, store_metadata_data, metadata_context):
        data = dict()
        if hasattr(resource, "id"):
            data["@id"] = resource.id
            if store_metadata:
                data["@context"] = _merge_jsonld(resource.context, metadata_context.iri)
            else:
                data["@context"] = resource.context
        data.update({
            "@type": resource.type,
            "@id": resource.id,
            "name": resource.name
        })
        if hasattr(resource, "id") and store_metadata:
            data.update(store_metadata_data)
        return data
    return _make_jsonld_compacted




#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import numpy as np
import pytest
from pandas import DataFrame


# Test suite for conversion of resource to / from Pandas DataFrame.

# FIXME "expanded" not yet implemented because not yet done in as_json() (DKE-130)


@pytest.fixture
def df():
    return DataFrame({
        "id": {0: "123", 1: "345"},
        "type": {0: "Type", 1: "Type"},
        "p1": {0: "v1a", 1: "v1b"},
        "p2": {0: "v2a", 1: "v2b"}
    })


@pytest.fixture
def ndf():
    return DataFrame({
        "id": {0: "678", 1: "912"},
        "type": {0: "Other", 1: "Other"},
        "p3": {0: "v3c", 1: "v3d"},
        "p4.id": {0: "123", 1: "345"},
        "p4.type": {0: "Type", 1: "Type"},
        "p4.p1": {0: "v1a", 1: "v1b"},
        "p4.p2": {0: "v2a", 1: "v2b"}
    })


class TestConversionToDataFrame:

    def test_as_dataframe_basic(self, forge, df, r1, r2):
        rcs = [r1, r2]
        x = forge.as_dataframe(rcs)
        assert x.equals(df)

    def test_as_dataframe_metadata(self, forge, df, r1, r2):
        rcs = [r1, r2]
        forge.register(rcs)
        x = forge.as_dataframe(rcs, store_metadata=True)
        df["deprecated"] = [False, False]
        df["version"] = [1, 1]
        assert x.equals(df)

    def test_as_dataframe_missing(self, forge, df, r1, r2):
        # A Resource is missing a property.
        del r2.p1
        rcs = [r1, r2]
        x = forge.as_dataframe(rcs)
        df["p1"][1] = np.nan
        assert x.equals(df)

    def test_as_dataframe_na_none(self, forge, df, r1, r2):
        # A Resource has one property with a value which represents a missing value.
        r2.p1 = None
        rcs = [r1, r2]
        x = forge.as_dataframe(rcs)
        df["p1"][1] = np.nan
        assert x.equals(df)

    def test_as_dataframe_na_string(self, forge, df, r1, r2):
        # A Resource has one property with a value which represents a missing value.
        r2.p1 = "(missing)"
        rcs = [r1, r2]
        x = forge.as_dataframe(rcs, na="(missing)")
        df["p1"][1] = np.nan
        assert x.equals(df)

    def test_as_dataframe_na_strings(self, forge, df, r1, r2):
        # A Resource has two properties with two different values which represent a missing value.
        r1.p2 = "NA"
        r2.p1 = "(missing)"
        rcs = [r1, r2]
        x = forge.as_dataframe(rcs, na=["NA", "(missing)"])
        df["p2"][0] = np.nan
        df["p1"][1] = np.nan
        assert x.equals(df)

    def test_as_dataframe_nesting_default_depth_1(self, ndf, forge, r3, r4):
        rcs = [r3, r4]
        x = forge.as_dataframe(rcs)
        assert x.equals(ndf)

    def test_as_dataframe_nesting_default_depth_2(self, forge, ndf, r1, r2, r3, r4, r5):
        r1.p2 = r5
        del r2.p2
        rcs = [r3, r4]
        x = forge.as_dataframe(rcs)
        ndf.drop("p4.p2", axis=1, inplace=True)
        ndf["p4.p2.p5"] = ["v5e", np.nan]
        ndf["p4.p2.p6"] = ["v6e", np.nan]
        assert x.equals(ndf)

    def test_as_dataframe_nesting_specific(self, forge, ndf, r3, r4):
        rcs = [r3, r4]
        x = forge.as_dataframe(rcs, nesting="__")
        ndf.rename(columns={
            "p4.id": "p4__id",
            "p4.type": "p4__type",
            "p4.p1": "p4__p1",
            "p4.p2": "p4__p2",
        }, inplace=True)
        assert x.equals(ndf)

    def test_as_dataframe_nested_missing(self, forge, ndf, r2, r3, r4):
        del r2.p1
        rcs = [r3, r4]
        x = forge.as_dataframe(rcs)
        ndf["p4.p1"][1] = np.nan
        assert x.equals(ndf)

    def test_as_dataframe_nested_na_none(self, forge, ndf, r2, r3, r4):
        r2.p1 = None
        rcs = [r3, r4]
        x = forge.as_dataframe(rcs)
        ndf["p4.p1"][1] = np.nan
        assert x.equals(ndf)

    def test_as_dataframe_nested_na_string(self, forge, ndf, r2, r3, r4):
        r2.p1 = "(missing)"
        rcs = [r3, r4]
        x = forge.as_dataframe(rcs, na="(missing)")
        ndf["p4.p1"][1] = np.nan
        assert x.equals(ndf)

    def test_as_dataframe_nested_na_strings(self, forge, ndf, r1, r2, r3, r4):
        r1.p2 = "NA"
        r2.p1 = "(missing)"
        rcs = [r3, r4]
        x = forge.as_dataframe(rcs, na=["NA", "(missing)"])
        ndf["p4.p2"][0] = np.nan
        ndf["p4.p1"][1] = np.nan
        assert x.equals(ndf)


class TestConversionFromDataFrame:

    def test_from_dataframe_basic_row(self, forge, df, r1):
        df.drop(1, inplace=True)
        x = forge.from_dataframe(df)
        assert x == r1

    def test_from_dataframe_basic_rows(self, forge, df, r1, r2):
        x = forge.from_dataframe(df)
        rcs = [r1, r2]
        assert x == rcs

    def test_from_dataframe_na_none(self, forge, df, r1, r2):
        # A cell has a value which represents a missing value.
        df["p1"][1] = None
        x = forge.from_dataframe(df)
        del r2.p1
        rcs = [r1, r2]
        assert x == rcs

    def test_from_dataframe_na_string(self, forge, df, r1, r2):
        # A cell has a value which represents a missing value.
        df["p1"][1] = "(missing)"
        x = forge.from_dataframe(df, na="(missing)")
        del r2.p1
        rcs = [r1, r2]
        assert x == rcs

    def test_from_dataframe_na_strings(self, forge, df, r1, r2):
        # Two cells have two different values which represent a missing value.
        df["p2"][0] = "NA"
        df["p1"][1] = "(missing)"
        x = forge.from_dataframe(df, na=["NA", "(missing)"])
        del r1.p2
        del r2.p1
        rcs = [r1, r2]
        assert x == rcs

    def test_from_dataframe_nesting_default_depth_1(self, ndf, forge, r3, r4):
        x = forge.from_dataframe(ndf)
        rcs = [r3, r4]
        assert x == rcs

    def test_from_dataframe_nesting_default_depth_2(self, forge, ndf, r1, r2, r3, r4, r5):
        ndf.drop("p4.p2", axis=1, inplace=True)
        ndf["p4.p2.p5"] = ["v5e", np.nan]
        ndf["p4.p2.p6"] = ["v6e", np.nan]
        x = forge.from_dataframe(ndf)
        r1.p2 = r5
        del r2.p2
        rcs = [r3, r4]
        assert x == rcs

    def test_from_dataframe_nesting_specific(self, forge, ndf, r3, r4):
        ndf.rename(columns={
            "p4.id": "p4__id",
            "p4.type": "p4__type",
            "p4.p1": "p4__p1",
            "p4.p2": "p4__p2",
        }, inplace=True)
        x = forge.from_dataframe(ndf, nesting="__")
        rcs = [r3, r4]
        assert x == rcs

    def test_from_dataframe_nesting_na_none(self, forge, ndf, r2, r3, r4):
        ndf["p4.p1"][1] = None
        x = forge.from_dataframe(ndf)
        del r2.p1
        rcs = [r3, r4]
        assert x == rcs

    def test_from_dataframe_nesting_na_string(self, forge, ndf, r2, r3, r4):
        ndf["p4.p1"][1] = "(missing)"
        x = forge.from_dataframe(ndf, na="(missing)")
        del r2.p1
        rcs = [r3, r4]
        assert x == rcs

    def test_from_dataframe_nesting_na_strings(self, forge, ndf, r1, r2, r3, r4):
        ndf["p4.p2"][0] = "NA"
        ndf["p4.p1"][1] = "(missing)"
        x = forge.from_dataframe(ndf, na=["NA", "(missing)"])
        del r1.p2
        del r2.p1
        rcs = [r3, r4]
        assert x == rcs



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import json

import pytest
from urllib.error import URLError

from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import _merge_jsonld


def test_load_context_as_dict(custom_context):
    context = Context(custom_context)
    assert is_valid_document(context.document)
    assert context.document == custom_context
    assert context.iri is None
    assert context.is_http_iri() is False


def test_load_context_from_file(context_file_path, context_iri_file):
    with open(context_file_path) as f:
        context_json = json.loads(f.read())
    context = Context(context_iri_file, context_iri_file)
    assert is_valid_document(context.document)
    assert context.document == context_json
    assert context.iri is context_iri_file
    assert context.is_http_iri() is False


def test_load_context_from_list(custom_context, context_iri_file, model_prefixes):
    test_context = _merge_jsonld(custom_context, context_iri_file)
    context = Context(test_context)
    assert is_valid_document(context.document)
    assert context.iri is None
    assert context.is_http_iri() is False
    assert context.base == "http://example.org/"
    assert context.vocab == "http://example.org/vocab/"
    assert context.expand("Person") == "http://schema.org/Person"
    assert context.prefixes == model_prefixes


def test_load_context_from_url():
    context_url = "https://json-ld.org/contexts/person.jsonld"
    context = Context(context_url, context_url)
    assert is_valid_document(context.document)
    assert context.expand("affiliation") == "http://schema.org/affiliation"
    assert context.iri == context_url
    assert context.is_http_iri() is True


def test_load_context_fail():
    context_url = "https://unresolvable.context.org"
    with pytest.raises(URLError):
        Context(context_url)


def is_valid_document(doc):
    return (isinstance(doc, dict)
            and "@context" in doc)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import pytest


# Test suite for conversion of a resource to / from JSON.

@pytest.fixture
def json_one():
    return {
        "id": "123",
        "type": "Type",
        "p1": "v1a",
        "p2": "v2a"
    }


@pytest.fixture
def json_many():
    return [
        {
            "id": "123",
            "type": "Type",
            "p1": "v1a",
            "p2": "v2a"
        },
        {
            "id": "345",
            "type": "Type",
            "p1": "v1b",
            "p2": "v2b"
        }
    ]


@pytest.fixture
def json_nested_many():
    return [
        {
            "id": "678",
            "type": "Other",
            "p3": "v3c",
            "p4": {
                "id": "123",
                "type": "Type",
                "p1": "v1a",
                "p2": "v2a"
            }
        },
        {
            "id": "912",
            "type": "Other",
            "p3": "v3d",
            "p4": {
                "id": "345",
                "type": "Type",
                "p1": "v1b",
                "p2": "v2b"
            }
        }
    ]


@pytest.fixture
def json_list_one():
    return {
        "id": "678",
        "type": "Other",
        "p3": "v3c",
        "p4": [
            {
                "id": "123",
                "type": "Type",
                "p1": "v1a",
                "p2": "v2a"
            },
            {
                "id": "345",
                "type": "Type",
                "p1": "v1b",
                "p2": "v2b"
            }
        ]
    }


class TestConversionFromJson:

    def test_from_json_basic_one(self, forge, json_one, r1):
        x = forge.from_json(json_one)
        assert x == r1

    def test_from_json_list_resources_one(self, forge, json_list_one, r1, r2, r3):
        x = forge.from_json(json_list_one)
        r3.p4 = [r1, r2]
        assert x == r3

    def test_from_json_list_mixed_one(self, forge, json_list_one, r1, r2, r3):
        json_list_one["p4"].insert(1, "543")
        x = forge.from_json(json_list_one)
        r3.p4 = [r1, "543", r2]
        assert x == r3

    def test_from_json_basic_many(self, forge, json_many, r1, r2):
        x = forge.from_json(json_many)
        rcs = [r1, r2]
        assert x == rcs

    def test_from_json_na_none(self, forge, json_many, r1, r2):
        json_many[1]["p1"] = None
        x = forge.from_json(json_many)
        del r2.p1
        rcs = [r1, r2]
        assert x == rcs

    def test_from_json_na_string(self, forge, json_many, r1, r2):
        json_many[1]["p1"] = "(missing)"
        x = forge.from_json(json_many, na="(missing)")
        del r2.p1
        rcs = [r1, r2]
        assert x == rcs

    def test_from_json_na_strings(self, forge, json_many, r1, r2):
        json_many[0]["p2"] = "NA"
        json_many[1]["p1"] = "(missing)"
        x = forge.from_json(json_many, na=["NA", "(missing)"])
        del r1.p2
        del r2.p1
        rcs = [r1, r2]
        assert x == rcs

    def test_from_json_nested_default_depth_1(self, json_nested_many, forge, r3, r4):
        x = forge.from_json(json_nested_many)
        rcs = [r3, r4]
        assert x == rcs

    def test_from_json_nested_default_depth_2(self, forge, json_nested_many, r1, r3, r4, r5):
        json_nested_many[0]["p4"]["p2"] = {"p5": "v5e", "p6": "v6e"}
        x = forge.from_json(json_nested_many)
        r1.p2 = r5
        rcs = [r3, r4]
        assert x == rcs

    def test_from_json_nested_na_none(self, forge, json_nested_many, r2, r3, r4):
        json_nested_many[1]["p4"]["p1"] = None
        x = forge.from_json(json_nested_many)
        del r2.p1
        rcs = [r3, r4]
        assert x == rcs

    def test_from_json_nested_na_string(self, forge, json_nested_many, r2, r3, r4):
        json_nested_many[1]["p4"]["p1"] = "(missing)"
        x = forge.from_json(json_nested_many, na="(missing)")
        del r2.p1
        rcs = [r3, r4]
        assert x == rcs

    def test_from_json_nested_na_strings(self, forge, json_nested_many, r1, r2, r3, r4):
        json_nested_many[0]["p4"]["p2"] = "NA"
        json_nested_many[1]["p4"]["p1"] = "(missing)"
        x = forge.from_json(json_nested_many, na=["NA", "(missing)"])
        del r1.p2
        del r2.p1
        rcs = [r3, r4]
        assert x == rcs



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import KeysView, Set, Tuple


def check_collisions(reserved: Set[str], new: KeysView[str]) -> None:
    intersect = reserved.intersection(set(new))
    if intersect:
        raise NotImplementedError(f"some names of the given properties are reserved: {intersect}")


def sort_attrs(kv: Tuple[str, str]) -> Tuple[int, str]:
    # POLICY Should be called to sort attributes of resources, templates, mappings, ...
    ordered = ["_last_action", "_validated", "_synchronized", "_store_metadata",
               "context", "id", "type", "label"]
    orders = {x: i for i, x in enumerate(ordered)}
    next_order = len(ordered) + 1
    return orders.get(kv[0], next_order), kv[0]


def repr_class(self: object) -> str:
    ordered = sorted(self.__dict__.items(), key=sort_attrs)
    attributes = (f"{k}={v!r}" for k, v in ordered)
    attributes_str = ", ".join(attributes)
    class_name = type(self).__name__
    return f"{class_name}({attributes_str})"


def eq_class(self: object, other: object) -> bool:
    return self.__dict__ == other.__dict__ if type(other) is type(self) else False



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from collections import Counter
from typing import Any, Callable, Iterable, Iterator, List, Optional, Union

from kgforge.core import Resource
from kgforge.core.commons.attributes import eq_class, repr_class


class Action:

    def __init__(self, operation: str, succeeded: bool, error: Optional[Exception]) -> None:
        self.operation: str = operation
        self.succeeded: bool = succeeded
        self.error: Optional[str] = type(error).__name__ if error else None
        self.message: Optional[str] = str(error) if error else None

    def __repr__(self) -> str:
        return repr_class(self)

    def __str__(self) -> str:
        error = f"\n<error> {self.error}: {self.message}" if self.error is not None else ""
        return (f"<action> {self.operation}"
                f"\n<succeeded> {self.succeeded}"
                f"{error}")

    def __eq__(self, other: object) -> bool:
        return eq_class(self, other)

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.__dict__.items())))


class Actions(list):

    def __init__(self, actions: Iterable[Action]) -> None:
        super().__init__(actions)

    # No need to define __repr__ as the one from list are suitable.

    def __str__(self) -> str:
        counted = Counter(self)
        return "\n\n".join(f"<count> {count}\n{action}" for action, count in counted.items())

    @staticmethod
    def from_resources(resources: List[Resource]):
        return Actions(x._last_action for x in resources)


class LazyAction:

    def __init__(self, operation: Callable, *args) -> None:
        self.operation: Callable = operation
        self.args = args

    def __repr__(self) -> str:
        return repr_class(self)

    def __str__(self) -> str:
        return f"LazyAction(operation={self.operation.__qualname__}, args={list(self.args)})"

    def __eq__(self, other: object) -> bool:
        return eq_class(self, other)

    def execute(self) -> Union[Resource, List[Resource]]:
        # POLICY Operation should propagate exceptions. This is for actions.execute_lazy_actions().
        return self.operation(*self.args)


def execute_lazy_actions(resource: Resource, lazy_actions: List[str]) -> None:
    # TODO Use as base an implementation of JSONPath for Python. DKE-147.
    for path in lazy_actions:
        lazy_action = eval(path, {}, {"x": resource})
        result = lazy_action.execute()
        rc_path, rc_prop = path.rsplit(".", maxsplit=1)
        rc = eval(rc_path, {}, {"x": resource})
        if not rc_prop.endswith("]"):
            setattr(rc, rc_prop, result)
        else:
            exec(f"{path} = result", {}, {"x": resource, "result": result})


def collect_lazy_actions(resource: Resource) -> List[str]:
    return list(_collect_lazy_actions(resource, "x"))


def _collect_lazy_actions(data: Any, root: str) -> Iterator[str]:
    if isinstance(data, Resource):
        for k, v in data.__dict__.items():
            path = f"{root}.{k}"
            if isinstance(v, List):
                for i, x in enumerate(v):
                    if isinstance(x, LazyAction):
                        yield f"{path}[{i}]"
                    else:
                        yield from _collect_lazy_actions(x, f"{path}[{i}]")
            elif isinstance(v, Resource):
                yield from _collect_lazy_actions(v, path)
            elif isinstance(v, LazyAction):
                yield path



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the test suite for actions.
import pytest

from kgforge.core.archetypes.store import rewrite_sparql


context = {
    "@context": {
        "type": {
            "@id": "rdf:type",
            "@type": "@id"
        },
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "prov": "http://www.w3.org/ns/prov#",
        "schema": "http://schema.org/",
        "Person": {
            "@id": "schema:Person",
            "@type": "@id"
        },
        "Association": "prov:Association",
        "name": "schema:name",
        "agent": "prov:agent",
        "description": "http://schema.org/description",
    }
}


prefixes = {
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "prov": "http://www.w3.org/ns/prov#",
    "schema": "http://schema.org/",
}


prefixes_string = "\n".join([f"PREFIX {k}: <{v}>" for k, v in prefixes.items()])


form_store_metadata_combinations = [
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> agent ?agent }",
     prefixes_string + "\nSELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }"),
    ("SELECT ?agent WHERE { ?agent type ?v0 FILTER(?v0 != Person) }",
     prefixes_string + "\nSELECT ?agent WHERE { ?agent rdf:type ?v0 FILTER(?v0 != schema:Person) }"),
    ("SELECT ?x ?name WHERE { ?x type Association ; agent/name ?name }",
     prefixes_string + "\nSELECT ?x ?name WHERE { ?x rdf:type prov:Association ; prov:agent/schema:name ?name }"),
    ("SELECT ?name WHERE { ?x agent/name ?name FILTER regex(?name, \"^j\", \"i\") }",
     prefixes_string + "\nSELECT ?name WHERE { ?x prov:agent/schema:name ?name FILTER regex(?name, \"^j\", \"i\") }"),
    ("SELECT ?x WHERE { <http://exaplpe.org/1234> description ?x }",
     prefixes_string + "\nSELECT ?x WHERE { <http://exaplpe.org/1234> <http://schema.org/description> ?x }"),
]


@pytest.mark.parametrize("query, expected", form_store_metadata_combinations)
def test_rewrite_sparql(query, expected):
    result = rewrite_sparql(query, context, prefixes)
    assert result == expected


#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from abc import ABC, abstractmethod
from typing import List, Tuple, Dict, Optional, Set

from pyshacl.constraints.core.cardinality_constraints import SH_minCount
from pyshacl.constraints.core.logical_constraints import SH_and, SH_or, SH_xone
from pyshacl.constraints.core.other_constraints import SH_in, SH_hasValue
from pyshacl.constraints.core.value_constraints import SH_class, SH_nodeKind, SH_datatype
from pyshacl.consts import SH_property, SH_node, SH_IRI, SH_BlankNodeOrIRI, SH_targetClass
from pyshacl.shape import Shape
from rdflib import RDF
from rdflib.term import URIRef, BNode

from kgforge.specializations.models.rdf.node_properties import NodeProperties


ID_URI = URIRef('id')


class Collector(ABC):
    """Collector abstract class

    Depending on the constraint, a set of properties and attributes may be
    collected through the collect function.

    Attributes:
        shape: the Shacl Shape used by the collector
    """

    def __init__(self, shape: Shape) -> None:
        self.shape = shape

    @classmethod
    @abstractmethod
    def constraint(cls) -> URIRef:
        """Returns the Shacl constraint URI of the collector"""
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def collect(cls, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                         Optional[Dict]]:
        """ depending on the constraint, this function will return properties, attributes or
        both.

        Args:
            predecessors: list of nodes that have being traversed, used to break circular
                recursion

        Returns:
            properties, attributes: Tuple(list,dict), the collected properties and attributes
                respectively
        """
        raise NotImplementedError()

    def get_shape_target_classes(self) -> List:
        """Returns a list of target and implicit classes if any of the shape

        Returns:
            list of URIs
        """
        (_, target_classes, implicit_classes, _, _) = self.shape.target()
        target_classes = set(target_classes)
        target_classes.update(set(implicit_classes))
        return list(target_classes)


class HasValueCollector(Collector):
    """This class will collect values in the sh:hasValue constraint as attribute"""
    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.has_value = next(shape.objects(SH_hasValue))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_hasValue

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        attrs = dict()
        attrs["values"] = self.has_value
        return None, attrs


class DatatypeCollector(Collector):
    """This class will collect values in the sh:datatype constraint as attribute"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.data_type_rule = next(self.shape.objects(SH_datatype))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_datatype

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        attrs = dict()
        attrs["values"] = self.data_type_rule
        return None, attrs


class MinCountCollector(Collector):
    """This class will collect values in the sh:minCount constraint as attribute"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.min_count = next(self.shape.objects(SH_minCount))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_minCount

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        attrs = dict()
        attrs["mandatory"] = True if self.min_count.toPython() >= 1 else False
        return None, attrs


class NodeKindCollector(Collector):
    """This class will collect values in the sh:nodeKind constraint as attribute"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.node_kind_rule = next(self.shape.objects(SH_nodeKind))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_nodeKind

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        properties = list()
        if self.node_kind_rule == SH_IRI:
            properties.append(id_node_property(True))
        elif self.node_kind_rule == SH_BlankNodeOrIRI:
            properties.append(id_node_property(False))

        return properties, None


class InCollector(Collector):
    """This class will collect values in the sh:in constraint as attribute"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        sg = self.shape.sg.graph
        self.in_list = next(self.shape.objects(SH_in))
        self.in_values = set(sg.items(self.in_list))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_in

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        attrs = dict()
        attrs.update(constraint="in")
        attrs.update(values=[v.toPython() for v in self.in_values])
        return None, attrs


class ClassCollector(Collector):
    """This class will collect values in the sh:class constraint as attribute"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.class_rules = list(self.shape.objects(SH_class))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_class

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        attributes = dict()
        properties = list()
        for target_class in self.class_rules:
            target_class_shapes = [s for s in self.shape.sg.graph.subjects(SH_targetClass, target_class)]
            if target_class_shapes:
                for target_class_shape in target_class_shapes:
                    target_shape = self.shape.get_other_shape(target_class_shape)
                    if target_shape.node not in predecessors:
                        props, attrs = target_shape.traverse(predecessors)
                        if props:
                            attrs.update(properties=props)
                        properties.append(NodeProperties(**attrs))
                    else:
                        properties.append(type_node_property(target_class, True))
                        properties.append(id_node_property(True))
            else:
                properties.append(id_node_property(True))
                properties.append(type_node_property(target_class, True))

        # TODO: if we want not to navigate into the Class, use only attributes instead of
        #  the above for loop
        # attributes = {
        #     "path": RDF.type,
        #     "values":  [v for v in self.class_rules]
        # }
        return properties, attributes


class NodeCollector(Collector):
    """This class will collect values in the sh:node constraint as properties or
    attributes"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.node_shapes = list(self.shape.objects(SH_node))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_node

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        properties = list()
        attributes = dict()
        for n_shape in self.node_shapes:
            ns = self.shape.get_other_shape(n_shape)
            if ns.node not in predecessors:
                props, attrs = ns.traverse(predecessors)
                if ns.path() is not None:
                    if not isinstance(ns.path(), BNode):
                        attrs["path"] = ns.path()
                        if props:
                            attrs["properties"] = props
                        p = NodeProperties(**attrs)
                        properties.append(p)
                else:
                    properties.extend(props)
                    attributes.update(attrs)

        return properties, attributes


class PropertyCollector(Collector):
    """This class will collect values in the sh:property constraint as properties or
        attributes"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.property_shapes = list(self.shape.objects(SH_property))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_property

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        properties = list()
        for p_shape in self.property_shapes:
            ps = self.shape.get_other_shape(p_shape)
            if ps.node not in predecessors:
                props, attrs = ps.traverse(predecessors)
                if ps.path() is not None:
                    if not isinstance(ps.path(), BNode):
                        attrs["path"] = ps.path()
                        if props:
                            attrs["properties"] = props
                        p = NodeProperties(**attrs)
                        properties.append(p)

        types = self.get_shape_target_classes()
        if len(types) > 0:
            properties.append(type_node_property(types, True))

        return properties, None


class AndCollector(Collector):
    """This class will collect values in the sh:and constraint as properties"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.and_list = list(self.shape.objects(SH_and))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_and

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        properties = list()
        sg = self.shape.sg.graph
        for and_c in self.and_list:
            and_list = set(sg.items(and_c))
            for and_shape in and_list:
                and_shape = self.shape.get_other_shape(and_shape)
                if and_shape.node not in predecessors:
                    p, a = and_shape.traverse(predecessors)
                    if a is not None:
                        if and_shape.path() is not None:
                            if not isinstance(and_shape.path(), BNode):
                                a["path"] = and_shape.path()
                        if len(p) > 0:
                            a["properties"] = p
                        node = NodeProperties(**a)
                        properties.append(node)
                    else:
                        properties.extend(p)
        types = self.get_shape_target_classes()
        if len(types) > 0:
            properties.append(type_node_property(types, True))
        return properties, None


class OrCollector(Collector):
    """This class will collect values in the sh:and constraint as properties"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.or_list = list(self.shape.objects(SH_or))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_or

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        properties = list()
        attributes = dict()
        sg = self.shape.sg.graph
        for or_c in self.or_list:
            or_list = set(sg.items(or_c))
            for or_shape in or_list:
                or_shape = self.shape.get_other_shape(or_shape)
                if or_shape.node not in predecessors:
                    props, attrs = or_shape.traverse(predecessors)
                    if or_shape.path() is not None:
                        # This Node concerns PropertyNode options
                        if not isinstance(or_shape.path(), BNode):
                            attrs["path"] = or_shape.path()
                            if props:
                                attrs["properties"] = props
                            node = NodeProperties(**attrs)
                            properties.append(node)
                    else:
                        # This concerns ShapeNode options
                        if props:
                            properties.extend(props)
                            if len(properties) > 1:
                                mandatory_ids = get_nodes_path(properties, ID_URI, "mandatory")
                                types = get_nodes_path(properties, RDF.type, "values")
                                properties.clear()
                                if mandatory_ids:
                                    is_mandatory = any(x for x in mandatory_ids)
                                    properties.append(id_node_property(is_mandatory))
                                if types:
                                    properties.append(type_node_property(types, True))
                        elif attrs:
                            attributes = merge_dicts(attributes, attrs)
        types = self.get_shape_target_classes()
        if len(types) > 0:
            properties.append(type_node_property(types, True))
        attributes["constraint"] = "or"
        return properties, attributes


class XoneCollector(Collector):
    """This class will collect values in the sh:and constraint as properties"""

    def __init__(self, shape: Shape) -> None:
        super().__init__(shape)
        self.xone_list = list(self.shape.objects(SH_xone))

    @classmethod
    def constraint(cls) -> URIRef:
        return SH_xone

    def collect(self, predecessors: Set[URIRef]) -> Tuple[Optional[List[NodeProperties]],
                                                          Optional[Dict]]:
        properties = list()
        attributes = dict()
        sg = self.shape.sg.graph

        for xone_c in self.xone_list:
            xone_list = set(sg.items(xone_c))
            for xone_shape in xone_list:
                xone_shape = self.shape.get_other_shape(xone_shape)
                if xone_shape.node not in predecessors:
                    props, attrs = xone_shape.traverse(predecessors)
                    if xone_shape.path() is not None:
                        # This Node concerns PropertyNode options
                        if not isinstance(xone_shape.path(), BNode):
                            attrs["path"] = xone_shape.path()
                            if props:
                                attrs["properties"] = props
                            node = NodeProperties(**attrs)
                            properties.append(node)
                    else:
                        # This concerns ShapeNode options
                        if props:
                            properties.extend(props)
                            if len(properties) > 1:
                                mandatory_ids = get_nodes_path(properties, ID_URI, "mandatory")
                                types = get_nodes_path(properties, RDF.type, "values")
                                properties.clear()
                                if mandatory_ids:
                                    is_mandatory = any(x for x in mandatory_ids)
                                    properties.append(id_node_property(is_mandatory))
                                if types:
                                    properties.append(type_node_property(types, False))
                        elif attrs:
                            attributes = merge_dicts(attributes, attrs)
        types = self.get_shape_target_classes()
        if len(types) > 0:
            properties.append(type_node_property(types, True))

        attributes["constraint"] = "xone"
        return properties, attributes


# TODO: create some factories for NodeProperties

def type_node_property(types, mandatory: bool) -> NodeProperties:
    attrs = {"path": RDF.type, "values": types, "mandatory": mandatory}
    return NodeProperties(**attrs)


def id_node_property(mandatory: bool) -> NodeProperties:
    attrs = {"path": ID_URI, "mandatory": mandatory}
    return NodeProperties(**attrs)


def merge_dicts(first: Dict, second: Dict):
    result = dict()
    result.update({k: first[k] for k in first.keys() - second.keys()})
    result.update({k: second[k] for k in second.keys() - first.keys()})
    same = first.keys() & second.keys()
    for k in same:
        if str(first[k]) == str(second[k]):
            result[k] = first[k]
        else:
            if isinstance(first[k], list):
                if isinstance(second[k], list):
                    result[k] = list(set(first[k]) | set(second[k]))
                else:
                    result[k] = first[k]
                    result[k].append(second[k])
            else:
                if isinstance(second[k], list):
                    result[k] = second[k]
                    result[k].append(first[k])
                else:
                    result[k] = [first[k], second[k]]
    return result


def get_nodes_path(nodes: List, path: URIRef, field: str):
    if len(nodes) == 0:
        return None, []
    elif len(nodes) == 1:
        types = get_node_path(nodes[0], path, field)
    else:
        types = get_node_path(nodes[0], path, field)
        for node in nodes[1:]:
            second_types = get_node_path(node, path, field)
            types = types + second_types
    return types


def get_node_path(node: NodeProperties, path: URIRef, field: str):
    result = list()
    if hasattr(node, "properties"):
        for pp in node.properties:
            if hasattr(pp, "path"):
                if pp.path == path:
                    values = getattr(pp, field, None)
                    if isinstance(values, list):
                        result.extend(values)
                    else:
                        result.append(values)
    return result



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from kgforge.core import Resource
from kgforge.core.archetypes import Mapper, Mapping
from kgforge.core.conversions.json import from_json
from kgforge.core.wrappings.dict import DictWrapper, wrap_dict


# NB: Do not 'from kgforge.core import KnowledgeGraphForge' to avoid cyclic dependency.


class DictionaryMapper(Mapper):

    def __init__(self, forge: Optional["KnowledgeGraphForge"] = None) -> None:
        super().__init__(forge)

    def _map_one(self, data: Union[Path, Dict], mappings: List[Mapping], nas: List[Any]
                 ) -> List[Resource]:
        variables = {
            "forge": self.forge,
            "x": self._load_one(data),
        }
        mapped = (_apply_rules(x.rules, variables) for x in mappings)
        return [from_json(x, nas) for x in mapped]

    @staticmethod
    def _load_one(data: Union[Path, Dict]) -> DictWrapper:
        if isinstance(data, Path):
            with data.open() as f:
                record = json.load(f)
                return wrap_dict(record)
        else:
            return wrap_dict(data)


def _apply_rules(value: Any, variables: Dict) -> Union[Dict, List[Dict]]:
    if isinstance(value, Dict):
        return {k: _apply_rules(v, variables) for k, v in value.items()}
    elif isinstance(value, List):
        return [_apply_rules(x, variables) for x in value]
    else:
        # TODO Add support for the full syntax of JSONPath. DKE-147.
        try:
            return eval(value, variables, variables)
        except (TypeError, NameError, SyntaxError):
            return value



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import asyncio
import json
import re
from asyncio import Task
from collections import namedtuple
from copy import deepcopy
from enum import Enum
from typing import Callable, Dict, List, Optional, Union
from urllib.error import URLError
from urllib.parse import quote_plus

import nest_asyncio
import nexussdk as nexus
from aiohttp import ClientSession, hdrs

from kgforge.core import Resource
from kgforge.core.commons.actions import Action, collect_lazy_actions, execute_lazy_actions
from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import (_from_jsonld_one, _remove_ld_keys, as_jsonld,
                                          recursive_resolve)
from kgforge.core.wrappings.dict import wrap_dict


class BatchAction(Enum):
    CREATE = "create"
    FETCH = "fetch"
    DEPRECATE = "deprecate"
    UPDATE = "update"
    TAG = "tag"
    UPLOAD = "upload"
    DOWNLOAD = "download"


BatchResult = namedtuple("BatchResult", ["resource", "response"])
BatchResults = List[BatchResult]

NEXUS_NAMESPACE = "https://bluebrain.github.io/nexus/vocabulary/"
NEXUS_CONTEXT = "https://bluebrain.github.io/nexus/contexts/resource.json"
DEPRECATED_PROPERTY = f"{NEXUS_NAMESPACE}deprecated"


class Service:

    def __init__(self, endpoint: str, org: str, prj: str, token: str, model_context: Context,
                 max_connections: int):

        nexus.config.set_environment(endpoint)
        nexus.config.set_token(token)
        self.organisation = org
        self.project = prj
        self.model_context = model_context
        self.context = Context(self.get_project_context())
        self.context_cache: Dict = dict()
        self.metadata_context = Context(self.resolve_context(NEXUS_CONTEXT), NEXUS_CONTEXT)
        self.max_connections = max_connections
        self.headers = {
            "Authorization": "Bearer " + token,
            "Content-Type": "application/ld+json",
            "Accept": "application/ld+json"
        }
        self.headers_sparql = {
            "Authorization": "Bearer " + token,
            "Content-Type": "application/sparql-query",
            "Accept": "application/ld+json"
        }
        self.headers_upload = {
            "Authorization": "Bearer " + token,
            "Accept": "application/ld+json",
        }
        self.headers_download = {
            "Authorization": "Bearer " + token,
            "Accept": "*/*"
        }
        self.url_resources = "/".join((endpoint, "resources", quote_plus(org), quote_plus(prj)))
        self.url_files = "/".join((endpoint, "files", quote_plus(org), quote_plus(prj)))
        self.sparql_endpoint = "/".join((endpoint, "views", quote_plus(org), quote_plus(prj),
                                    "nxv:defaultSparqlIndex", "sparql"))
        # This async to work on jupyter notebooks
        nest_asyncio.apply()

    def get_project_context(self) -> Dict:
        project_data = nexus.projects.fetch(self.organisation, self.project)
        context = {
            "@base": project_data["base"],
            "@vocab": project_data["vocab"]
        }
        return context

    def resolve_context(self, iri: str, local_only: Optional[bool] = False) -> Dict:
        if iri in self.context_cache:
            return self.context_cache[iri]
        try:
            resource = nexus.resources.fetch(self.organisation, self.project, iri)
        except nexus.HTTPError:
            if local_only is False:
                try:
                    context = Context(iri)
                except URLError:
                    raise ValueError(f"{iri} is not resolvable")
                else:
                    document = context.document["@context"]
            else:
                raise ValueError(f"{iri} is not resolvable")
        else:
            document = json.loads(json.dumps(resource["@context"]))
        self.context_cache.update({iri: document})
        return document

    def batch_request(self, resources: List[Resource], action: BatchAction, callback: Callable,
                      error_type: Callable, **kwargs) -> (BatchResults, BatchResults):

        def create_tasks(semaphore, session, loop, data, batch_action, f_callback, error):
            futures = []
            for resource in data:
                if batch_action == batch_action.CREATE:
                    context = self.model_context or self.context
                    payload = as_jsonld(resource, "compacted", False,
                                        model_context=context, metadata_context=None,
                                        context_resolver=self.resolve_context)
                    schema_id = kwargs.get("schema_id")
                    schema_id = "_" if schema_id is None else quote_plus(schema_id)
                    url = f"{self.url_resources}/{schema_id}"
                    prepared_request = loop.create_task(
                        queue(hdrs.METH_POST, semaphore, session, url, resource, error, payload))

                if batch_action == batch_action.UPDATE:
                    url = "/".join((self.url_resources, "_", quote_plus(resource.id)))
                    params = {"rev": resource._store_metadata._rev}
                    payload = as_jsonld(resource, "compacted", False,
                                        model_context=self.model_context,
                                        metadata_context=None,
                                        context_resolver=self.resolve_context)
                    prepared_request = loop.create_task(
                        queue(hdrs.METH_PUT, semaphore, session, url, resource, error, payload,
                              params))

                if batch_action == batch_action.TAG:
                    url = "/".join((self.url_resources, "_", quote_plus(resource.id), "tags"))
                    rev = resource._store_metadata._rev
                    params = {"rev": rev}
                    payload = {"tag": kwargs.get("tag"), "rev": rev}
                    prepared_request = loop.create_task(
                        queue(hdrs.METH_POST, semaphore, session, url, resource, error, payload,
                              params))

                if batch_action == batch_action.DEPRECATE:
                    url = "/".join((self.url_resources, "_", quote_plus(resource.id)))
                    params = {"rev": resource._store_metadata._rev}
                    prepared_request = loop.create_task(
                        queue(hdrs.METH_DELETE, semaphore, session, url, resource, error,
                              params=params))

                if batch_action == BatchAction.FETCH:
                    url = "/".join((self.url_resources, "_", quote_plus(resource.id)))
                    prepared_request = loop.create_task(
                        queue(hdrs.METH_GET, semaphore, session, url, resource, error))

                if f_callback:
                    prepared_request.add_done_callback(f_callback)
                futures.append(prepared_request)
            return futures

        async def queue(method, semaphore, session, url, resource, exception, payload=None,
                        params=None):
            async with semaphore:
                return await request(method, session, url, resource, payload, params, exception)

        async def request(method, session, url, resource, payload, params, exception):
            async with session.request(method, url, headers=self.headers,
                                       data=json.dumps(payload), params=params) as response:
                content = await response.json()
                if response.status < 400:
                    return BatchResult(resource, content)
                else:
                    msg = " ".join(re.findall('[A-Z][^A-Z]*', content["@type"])).lower()
                    error = exception(msg)
                    return BatchResult(resource, error)

        async def dispatch_action():
            semaphore = asyncio.Semaphore(self.max_connections)
            loop = asyncio.get_event_loop()
            async with ClientSession() as session:
                tasks = create_tasks(semaphore, session, loop, resources, action, callback, error_type)
                return await asyncio.gather(*tasks)

        return asyncio.run(dispatch_action())

    def sync_metadata(self, resource: Resource, result: Dict) -> None:
        metadata = {"id": resource.id}
        keys = sorted(self.metadata_context.terms.keys())
        only_meta = {k: v for k, v in result.items() if k in keys}
        metadata.update(_remove_ld_keys(only_meta, self.metadata_context, False))
        resource._store_metadata = wrap_dict(metadata)

    def synchronize_resource(self, resource: Resource, response: Union[Exception, Dict],
                             action_name: str, succeeded: bool, synchronized: bool) -> None:
        if succeeded:
            action = Action(action_name, succeeded, None)
            self.sync_metadata(resource, response)
        else:
            action = Action(action_name, succeeded, response)
        resource._last_action = action
        resource._synchronized = synchronized

    def default_callback(self, fun_name: str) -> Callable:
        def callback(task: Task):
            result = task.result()
            if isinstance(result.response, Exception):
                self.synchronize_resource(
                    result.resource, result.response, fun_name, False, False)
            else:
                self.synchronize_resource(
                    result.resource, result.response, fun_name, True, True)
        return callback

    def verify(self, resources: List[Resource], function_name, exception: Callable,
               id_required: bool, required_synchronized: bool, execute_actions: bool) -> List[Resource]:
        valid = list()
        for resource in resources:
            if id_required and not hasattr(resource, "id"):
                error = exception("resource should have an id")
                self.synchronize_resource(resource, error, function_name, False, False)
                continue
            if required_synchronized is not None:
                synchronized = resource._synchronized
                if synchronized is not required_synchronized:
                    be_or_not_be = "be" if required_synchronized is True else "not be"
                    error = exception(f"resource should {be_or_not_be} synchronized")
                    self.synchronize_resource(resource, error, function_name, False, False)
                    continue
            if execute_actions:
                lazy_actions = collect_lazy_actions(resource)
                if lazy_actions is not None:
                    try:
                        execute_lazy_actions(resource, lazy_actions)
                    except Exception as e:
                        self.synchronize_resource(
                            resource, exception(e), function_name, False, False)
                        continue
            valid.append(resource)
        return valid

    def to_resource(self, payload: Dict) -> Resource:
        data_context = deepcopy(payload["@context"])
        data_context.remove(NEXUS_CONTEXT)
        data_context = data_context[0] if len(data_context) == 1 else data_context
        metadata = dict()
        data = dict()
        for k, v in payload.items():
            if k in self.metadata_context.terms.keys():
                metadata[k] = v
            else:
                data[k] = v

        if self.model_context and data_context == self.model_context.iri:
            resolved_ctx = self.model_context.document["@context"]
        else:
            resolved_ctx = recursive_resolve(data_context, self.resolve_context)
        data["@context"] = resolved_ctx
        resource = _from_jsonld_one(data)
        resource.context = data_context
        if len(metadata) > 0:
            self.sync_metadata(resource, metadata)
        return resource



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import Optional, Union, Dict, List

from rdflib_jsonld.context import Context as RdflibContext
from rdflib_jsonld.util import source_to_json


class Context(RdflibContext):
    """Context class will hold a JSON-LD context in two forms: iri and document.

    See also: https://w3c.github.io/json-ld-syntax/#the-context
    """

    def __init__(self, document: Union[Dict, List, str], iri: Optional[str] = None) -> None:
        """Initialize the Context and resolves the document if necessary.

        The document can be provided as a dictionary, list or string. If a dictionary or list is
        provided, we assume that nothing needs to be resolved. If a string is provided,
        it should correspond to a resolvable document (i.e. file://, http://) which will be
        resolved in the initialization, the result will become the document and the iri will
        be the initial document parameter. Will throw an exception if the context is not resolvable.

        Args:
            document (Dict, List, str): resolved or resolvable document
            iri (str): the iri for the provided document
        """
        super().__init__(document)
        if isinstance(document, list):
            sub_docs = dict()
            for x in document:
                sub_context = Context(x)
                sub_docs.update(sub_context.document["@context"])
            self.document = {"@context": sub_docs}
        elif isinstance(document, str):
            try:
                self.document = source_to_json(document)
            except Exception:
                raise ValueError("context not resolvable")
        elif isinstance(document, Dict):
            self.document = document if "@context" in document else {"@context": document}
        self.iri = iri
        self.prefixes = {v: k for k, v in self._prefixes.items() if k.endswith(("/", "#"))}

    def is_http_iri(self):
        if self.iri:
            return self.iri.startswith("http")
        else:
            return False



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the test suite for reshaping.
import pytest
from kgforge.core import Resource
from kgforge.core.reshaping import collect_values


def test_collect_values():
    simple = Resource(type="Experiment", url="file.gz")
    r = collect_values(simple, "url")
    assert simple.url in r, "url should be in the list"
    deep = Resource(type="Experiment", level1=Resource(level2=Resource(url="file.gz")))
    r = collect_values(deep, "level1.level2.url")
    assert deep.level1.level2.url in r, "url should be in the list"
    files = [Resource(type="Experiment", url=f"file{i}") for i in range(3)]
    r = collect_values(files, "url")
    assert ["file0", "file1", "file2"] == r, "three elements should be in the list"
    data_set = Resource(type="Dataset", hasPart=files)
    r = collect_values(data_set, "hasPart.url")
    assert ["file0", "file1", "file2"] == r, "three elements should be in the list"
    with pytest.raises(ValueError):
        collect_values(data_set, "fake.path", ValueError)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import json
from copy import deepcopy
from typing import Union, Dict, List, Tuple, Optional, Callable
from urllib.error import URLError, HTTPError

from enum import Enum
from pyld import jsonld
from rdflib import Graph

from kgforge.core.commons.actions import LazyAction
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import NotSupportedError
from kgforge.core.commons.execution import dispatch
from kgforge.core.resource import Resource


class Form(Enum):
    EXPANDED = "expanded"
    COMPACTED = "compacted"


def as_graph(data: Union[Resource, List[Resource]], store_metadata: bool,
             model_context: Optional[Context], metadata_context: Optional[Context],
             context_resolver: Optional[Callable]) -> Graph:
    return dispatch(data, _as_graph_many, _as_graph_one, store_metadata, model_context,
                    metadata_context, context_resolver)


def as_jsonld(data: Union[Resource, List[Resource]], form: str, store_metadata: bool,
              model_context: Optional[Context], metadata_context: Optional[Context],
              context_resolver: Optional[Callable]) -> Union[Dict, List[Dict]]:
    try:
        valid_form = Form(form.lower())
    except ValueError:
        supported_forms = tuple(item.value for item in Form)
        raise NotSupportedError(f"supported serialization forms are {supported_forms}")

    return dispatch(
        data, _as_jsonld_many, _as_jsonld_one, valid_form, store_metadata, model_context,
        metadata_context, context_resolver)


def from_jsonld(data: Union[Dict, List[Dict]]) -> Union[Resource, List[Resource]]:
    if isinstance(data, List) and all(isinstance(x, Dict) for x in data):
        return _from_jsonld_many(data)
    elif isinstance(data, Dict):
        return _from_jsonld_one(data)
    else:
        raise TypeError("not a dictionary nor a list of dictionaries")


def from_graph(data: Graph) -> Union[Resource, List[Resource]]:
    raise NotImplementedError("not implemented yet")


def _from_jsonld_many(dataset: List[Dict]) -> List[Resource]:
    return [_from_jsonld_one(data) for data in dataset]


def _from_jsonld_one(data: Dict) -> Resource:
    if "@context" in data:
        try:
            resolved_context = Context(data["@context"])
        except URLError:
            raise ValueError("context not resolvable")
        else:
            return _remove_ld_keys(data, resolved_context)
    else:
        raise NotImplementedError("not implemented yet (expanded json-ld)")


def _as_jsonld_many(resources: List[Resource], form: Form, store_metadata: bool,
                    model_context: Optional[Context], metadata_context: Optional[Context],
                    context_resolver: Optional[Callable]) -> List[Dict]:
    return [_as_jsonld_one(resource, form, store_metadata, model_context, metadata_context,
                           context_resolver)
            for resource in resources]


def _as_jsonld_one(resource: Resource, form: Form, store_metadata: bool,
                   model_context: Optional[Context], metadata_context: Optional[Context],
                   context_resolver: Optional[Callable]) -> Dict:

    context = _resource_context(resource, model_context, context_resolver)
    resolved_context = context.document
    output_context = context.iri if context.is_http_iri() else context.document["@context"]
    if store_metadata and resource._store_metadata:
        if metadata_context:
            resolved_context = _merge_jsonld(resolved_context["@context"],
                                             metadata_context.document["@context"])
            if metadata_context.is_http_iri():
                metadata_context_output = metadata_context.iri
            else:
                metadata_context_output = metadata_context.document["@context"]
            output_context = _merge_jsonld(output_context, metadata_context_output)
        else:
            raise NotSupportedError("no available context in the metadata")
    try:
        data_graph, metadata_graph = _as_graphs(resource, store_metadata, context, metadata_context)
    except Exception as e:
        raise ValueError(e)
    data_expanded = json.loads(data_graph.serialize(format="json-ld").decode("utf-8"))
    if hasattr(resource, "id"):
        if context.base:
            uri = context.resolve(resource.id)
        else:
            uri = str(data_graph.namespace_manager.absolutize(resource.id))
        frame = {"@id": uri}
    else:
        frame = dict()
        for k, v in resource.__dict__.items():
            if k not in Resource._RESERVED and not isinstance(v, (Resource, dict, list)):
                if k == "context":
                    continue
                elif k == "type":
                    t = context.expand(v)
                    if t:
                        frame["@type"] = context.expand(v)
                else:
                    key = context.expand(k)
                    if key and isinstance(v, str):
                        frame[key] = v
    data_framed = jsonld.frame(data_expanded, frame)
    if store_metadata is True and len(metadata_graph) > 0:
        metadata_expanded = json.loads(metadata_graph.serialize(format="json-ld").decode("utf-8"))
        metadata_framed = jsonld.frame(metadata_expanded, {"@id": resource.id})
        if form is Form.COMPACTED:
            data_compacted = jsonld.compact(data_framed, resolved_context)
            metadata_compacted = jsonld.compact(metadata_framed, resolved_context)
            metadata_compacted = dict(sorted(metadata_compacted.items()))
            data_and_meta = _merge_jsonld(data_compacted, metadata_compacted)
            data_and_meta["@context"] = output_context
            # this is to be able to register contexts documents
            if len(data_compacted) == 1 and hasattr(resource, "id"):
                data_compacted["@id"] = resource.id
            return data_and_meta
        elif form is Form.EXPANDED:
            data_expanded = _unpack_from_list(data_framed)

            metadata_expanded = _unpack_from_list(metadata_framed)
            metadata_expanded = dict(sorted(metadata_expanded.items()))
            return _merge_jsonld(data_expanded, metadata_expanded)
    else:
        if form is Form.COMPACTED:
            compacted = jsonld.compact(data_framed, resolved_context)
            compacted["@context"] = output_context
            # this is to be able to register contexts documents
            if len(compacted) == 1 and hasattr(resource, "id"):
                compacted["@id"] = resource.id
            return compacted
        elif form is Form.EXPANDED:
            return _unpack_from_list(data_framed)


def _as_graph_many(resources: List[Resource], store_metadata: bool, model_context: Optional[Context],
                   metadata_context: Optional[Context], context_resolver: Optional[Callable]) -> Graph:
    graph = Graph()
    for resource in resources:
        result = _as_graph_one(resource, store_metadata, model_context, metadata_context,
                               context_resolver)
        graph.parse(result)
    return graph


def _as_graph_one(resource: Resource, store_metadata: bool, model_context: Optional[Context],
                  metadata_context: Optional[Context], context_resolver: Optional[Callable]) -> Graph:
    json_ld = _as_jsonld_one(resource, Form.EXPANDED, store_metadata, model_context,
                             metadata_context, context_resolver)
    return Graph().parse(data=json.dumps(json_ld), format="json-ld")


def _as_graphs(resource: Resource, store_metadata: bool, context: Context,
               metadata_context: Context) -> Tuple[Graph, Graph]:
    """Returns a data and a metadata graph"""
    if hasattr(resource, "context"):
        output_context = resource.context
    else:
        output_context = context.iri if context.is_http_iri() else context.document["@context"]
    converted = _add_ld_keys(resource, output_context, context.base)
    converted["@context"] = context.document["@context"]
    return _dicts_to_graph(converted, resource._store_metadata, store_metadata, metadata_context)


def _dicts_to_graph(data: Dict, metadata: Dict, store_meta: bool,
                    metadata_context: Context) -> Tuple[Graph, Graph]:
    json_str = json.dumps(data)
    graph = Graph().parse(data=json_str, format="json-ld")
    meta_data_graph = Graph()
    if store_meta is True and metadata is not None:
        if "id" not in metadata:
            raise ValueError("no id in the metadata")
        metadata = _add_ld_keys(metadata, None, None)
        metadata["@context"] = metadata_context.document["@context"]
        try:
            meta_data_graph.parse(data=json.dumps(metadata), format="json-ld")
        except Exception:
            raise ValueError("generated an invalid json-ld")
    return graph, meta_data_graph


def recursive_resolve(context: Union[Dict, List, str], resolver: Optional[Callable]) -> Dict:
    document = dict()
    if isinstance(context, list):
        for x in context:
            document.update(recursive_resolve(x, resolver))
    elif isinstance(context, str):
        doc = resolver(context)
        document.update(recursive_resolve(doc, resolver))
    elif isinstance(context, dict):
        document.update(context)
    return document


def _resource_context(resource: Resource, model_context: Context, context_resolver: Callable) -> Context:
    if hasattr(resource, "context"):
        if model_context and resource.context == model_context.iri:
            context = model_context
        else:
            iri = resource.context if isinstance(resource.context, str) else None
            try:
                document = recursive_resolve(resource.context, context_resolver)
                context = Context(document, iri)
            except (HTTPError, URLError, NotSupportedError):
                try:
                    context = Context(resource.context, iri)
                except URLError:
                    raise ValueError(f"{resource.context} is not resolvable")
    else:
        context = model_context

    if context is None:
        raise NotSupportedError("no available context")

    return context


def _unpack_from_list(data):
    if isinstance(data, list):
        node = data
    elif isinstance(data, dict):
        if "@graph" in data:
            node = data["@graph"]
        else:
            return data
    else:
        return data
    if len(node) == 1:
        return node[0]
    else:
        return node


def _dicts_to_graph(data: Dict, metadata: Dict, store_meta: bool,
                    metadata_context: Context) -> Tuple[Graph, Graph]:
    json_str = json.dumps(data)
    graph = Graph().parse(data=json_str, format="json-ld")
    meta_data_graph = Graph()
    if store_meta is True and metadata is not None:
        if "id" not in metadata:
            raise ValueError("no id in the metadata")
        metadata = _add_ld_keys(metadata, None, None)
        metadata["@context"] = metadata_context.document["@context"]
        try:
            meta_data_graph.parse(data=json.dumps(metadata), format="json-ld")
        except Exception:
            raise ValueError("generated an invalid json-ld")
    return graph, meta_data_graph


def _add_ld_keys(rsc: Resource, context: Optional[Union[Dict, List, str]], base: Optional[str]) -> Dict:
    local_attrs = dict()
    ld_keys = {"id": "@id", "type": "@type"}
    local_context = None
    for k, v in rsc.__dict__.items():
        if k not in Resource._RESERVED:
            if k == "context":
                if v != context:
                    local_context = Context(v)
                    base = local_context.base
            else:
                key = ld_keys.get(k, k)
                if key == "@id" and local_context is not None:
                    local_attrs[key] = local_context.resolve(v)
                else:
                    if isinstance(v, Resource):
                        local_attrs[key] = _add_ld_keys(v, context, base)
                    elif isinstance(v, list):
                        local_attrs[key] = [_add_ld_keys(item, context, base)
                                            if isinstance(item, Resource) else item for item in v]
                    else:
                        if isinstance(v, LazyAction):
                            raise ValueError("can't convert, resource contains LazyActions")
                        local_attrs[key] = v.replace(base, "") if base and isinstance(v, str) else v
    return local_attrs


def _remove_ld_keys(dictionary: dict, context: Context,
                    to_resource: Optional[bool] = True) -> Union[Dict, Resource]:
    local_attrs = dict()
    for k, v in dictionary.items():
        if k == "@context":
            if v != context:
                local_attrs["context"] = v
        else:
            if k == "@id":
                local_attrs["id"] = context.resolve(v)
            elif k.startswith("@"):
                local_attrs[k[1:]] = v
            else:
                if isinstance(v, dict):
                    local_attrs[k] = _remove_ld_keys(v, context, to_resource)
                elif isinstance(v, list):
                    local_attrs[k] = [_remove_ld_keys(item, context, to_resource)
                                      if isinstance(item, dict) else item for item in v]
                else:
                    if k in context.terms:
                        if context.terms[k].type == "@id":
                            v = context.shrink_iri(v)
                    local_attrs[k] = v
    if to_resource:
        return Resource(**local_attrs)
    else:
        return local_attrs


# Context


def _merge_dict_into_list(list_: List, dictionary: Dict) -> List:
    merged = False
    for item in list_:
        if isinstance(item, dict):
            item.update(dictionary)
            merged = True
            break
    if not merged:
        list_.append(dictionary)


def _merge_str_into_list(list_: List, string: str) -> List:
    if string not in list_:
        list_.append(string)


def _merge_second_list_into_first(first: List, second: List) -> List:
    for element in second:
        if isinstance(element, str):
            _merge_str_into_list(first, element)
        elif isinstance(element, dict):
            _merge_dict_into_list(first, element)


def _merge_jsonld(first: Union[Dict, List, str], second: Union[Dict, List, str]) -> Union[Dict, List, str]:
    result = None
    if isinstance(first, str):
        if isinstance(second, str):
            if first != second:
                result = [first, second]
            else:
                result = first
        elif isinstance(second, dict):
            result = [first, second]
        elif isinstance(second, list):
            result = [first]
            for element in second:
                if element != first:
                    result.append(element)
    if isinstance(first, list):
        result = deepcopy(first)
        if isinstance(second, str):
            _merge_str_into_list(result, second)
        if isinstance(second, dict):
            _merge_dict_into_list(result, second)
        if isinstance(second, list):
            _merge_second_list_into_first(result, second)
    if isinstance(first, dict):
        if isinstance(second, str):
            result = [deepcopy(first), deepcopy(second)]
        elif isinstance(second, dict):
            result = deepcopy(first)
            result.update(second)
        elif isinstance(second, list):
            result = [deepcopy(first)]
            _merge_second_list_into_first(result, second)
    return result



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Test suite for functions on dictionaries.
from copy import copy

import pytest

from kgforge.core.commons.dictionaries import with_defaults


@pytest.fixture
def original():
    return {
        "k1": "v1a",
        "k2": "v2a",
        "k3": "v3a",
        "k4": "v4a",
    }


@pytest.fixture
def other():
    return {
        "k2": "v2b",
        "k3": "v3b",
        "k4": "v4b",
    }


class TestWithDefaults:

    # Same.

    def test_same_missing_missing(self, original, other):
        # Same Store name. Missing endpoint. Missing token.
        # => Use other endpoint and other token.
        original["k2"] = "v2b"
        del original["k3"]
        del original["k4"]
        with_defaults(original, other, "k2", "k2", ["k3", "k4"])
        assert original["k3"] == "v3b"
        assert original["k4"] == "v4b"

    def test_same_given_missing(self, original, other):
        # Same Store name. Given endpoint. Missing token.
        # => Use original endpoint and other token.
        original["k2"] = "v2b"
        del original["k4"]
        with_defaults(original, other, "k2", "k2", ["k3", "k4"])
        assert original["k3"] == "v3a"
        assert original["k4"] == "v4b"

    def test_same_given_given(self, original, other):
        # Same Store name. Given endpoint. Given token.
        # => Use original endpoint and original token.
        original["k2"] = "v2b"
        with_defaults(original, other, "k2", "k2", ["k3", "k4"])
        assert original["k3"] == "v3a"
        assert original["k4"] == "v4a"

    # Different.

    def test_different_name(self, original, other):
        # Different Store name. Do nothing
        original_copy = copy(original)
        with_defaults(original, other, "k2", "k2", ["k3", "k4"])
        assert original == original_copy



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from collections import OrderedDict

import hjson

from kgforge.core.archetypes.mapping import Mapping
from kgforge.core.commons.attributes import sort_attrs


class DictionaryMapping(Mapping):

    def __init__(self, mapping: str) -> None:
        super().__init__(mapping)

    @staticmethod
    def _load_rules(mapping: str) -> OrderedDict:
        return hjson.loads(mapping)

    @staticmethod
    def _normalize_rules(rules: OrderedDict) -> str:
        return hjson.dumps(rules, indent=4, item_sort_key=sort_attrs)



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import requests
from requests import RequestException

from kgforge.core.commons.attributes import repr_class


class Mapping(ABC):

    # See dictionaries.py in kgforge/specializations/mappings/ for a reference implementation.

    # POLICY Exceptions should not be catched so that the KnowledgeGraphForge initialization fails.
    # POLICY Methods of archetypes, except __init__, should not have optional arguments.

    # POLICY Implementations should be declared in kgforge/specializations/mappings/__init__.py.
    # POLICY Implementations should not add methods but private functions in the file.
    # TODO Create a generic parameterizable test suite for mappings. DKE-135.
    # POLICY Implementations should pass tests/specializations/mappings/test_mappings.py.

    def __init__(self, mapping: str) -> None:
        self.rules: Any = self._load_rules(mapping)

    def __repr__(self) -> str:
        return repr_class(self)

    def __str__(self):
        return self._normalize_rules(self.rules)

    def __eq__(self, other: object) -> bool:
        # FIXME To properly work the loading of rules should normalize them. DKE-184.
        # return eq_class(self, other)
        raise NotImplementedError

    @classmethod
    def load(cls, source: str):
        # source: Union[str, FilePath, URL].
        # Mappings could be loaded from a string, a file, or an URL.
        filepath = Path(source)
        if filepath.is_file():
            text = filepath.read_text()
        else:
            try:
                response = requests.get(source)
                response.raise_for_status()
                text = response.text
            except RequestException:
                text = source
        return cls(text)

    def save(self, path: str) -> None:
        # path: FilePath.
        normalized = self._normalize_rules(self.rules)
        filepath = Path(path)
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.write_text(normalized)

    @staticmethod
    @abstractmethod
    def _load_rules(mapping: str) -> Any:
        """Load the mapping rules according to there interpretation."""
        pass

    @staticmethod
    @abstractmethod
    def _normalize_rules(rules: Any) -> str:
        """Normalize the representation of the rules to compare saved mappings."""
        pass



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import pytest
from pytest_bdd import given, scenarios, then, when

from kgforge.core import Resource

# TODO To be port to the generic parameterizable test suite for resources in test_resources.py.
#  DKE-135.


scenarios("resource.feature")


@given("I create a resource with a property.")
def resource():
    return Resource(type="Entity")


@given("I create a resource with an other resource as property.")
def nresource():
    return Resource(type="Entity", contribution=Resource(type="Contribution"))


@when("I create a resource with a reserved attribute. Creation should fail.")
def reserved_attribute_error():
    with pytest.raises(NotImplementedError):
        Resource(_validated=True)


@then("I should be able to access it as an attribute.")
def access_property(resource):
    assert resource.type == "Entity"


@then("I should be able to access the nested resource properties as JSONPath.")
def access_nested_property(nresource):
    assert nresource.type == "Entity"
    assert nresource.contribution.type == "Contribution"



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Iterable, Iterator, List, Optional, Sequence, Union

from kgforge.core import Resource
from kgforge.core.archetypes import Mapping
from kgforge.core.commons.attributes import repr_class


# NB: Do not 'from kgforge.core import KnowledgeGraphForge' to avoid cyclic dependency.


class Mapper(ABC):

    # See dictionaries.py in kgforge/specializations/mappers/ for a reference implementation.

    # POLICY Methods of archetypes, except __init__, should not have optional arguments.

    # POLICY Implementations should be declared in kgforge/specializations/mappers/__init__.py.
    # POLICY Implementations should not add methods but private functions in the file.
    # TODO Create a generic parameterizable test suite for mappers. DKE-135.
    # POLICY Implementations should pass tests/specializations/mappers/test_mappers.py.

    def __init__(self, forge: Optional["KnowledgeGraphForge"] = None) -> None:
        self.forge: Optional["KnowledgeGraphForge"] = forge

    def __repr__(self) -> str:
        return repr_class(self)

    def map(self, data: Any, mapping: Union[Mapping, List[Mapping]], na: Union[Any, List[Any]]
            ) -> Union[Resource, List[Resource]]:
        # Data could be loaded from a directory, a file, a collection, or an object.
        mappings = mapping if isinstance(mapping, List) else [mapping]
        nas = na if isinstance(na, List) else [na]
        if isinstance(data, str):
            path = Path(data)
            if path.is_dir():
                paths = path.iterdir()
                mapped = self._map_many(paths, mappings, nas)
            else:
                mapped = self._map_one(path, mappings, nas)
        elif isinstance(data, (Sequence, Iterator)):
            # NB: Do not use 'Iterable' for checking type because a Dict is an Iterable.
            mapped = self._map_many(data, mappings, nas)
        else:
            mapped = self._map_one(data, mappings, nas)
        return mapped[0] if len(mapped) == 1 else mapped

    def _map_many(self, data: Iterable[Union[Path, Any]], mappings: List[Mapping], nas: List[Any]
                  ) -> List[Resource]:
        # Bulk mapping could be optimized by overriding this method in the specialization.
        # POLICY Should follow self._map_one() policies.
        # POLICY Should not separate records loading for proper parallel / distributed processing.
        return [y for x in data for y in self._map_one(x, mappings, nas)]

    @abstractmethod
    def _map_one(self, data: Union[Path, Any], mappings: List[Mapping], nas: List[Any]
                 ) -> List[Resource]:
        # Method to change to handle a new type of mapping to apply.
        # POLICY Should load the record only one time with self._load_one().
        # POLICY Should not include as Resource properties elements with values in 'na'.
        # POLICY Could make the KnowledgeGraphForge instance available as 'forge'.
        # POLICY Could make the record instance available as 'x'.
        pass

    # No _load_many() for proper parallel / distributed processing directly in self._map_many().

    @staticmethod
    @abstractmethod
    def _load_one(data: Union[Path, Any]) -> Any:
        # Method to change to handle a new type of data to map.
        # POLICY Should prepare the record according to how the rules are using it.
        # The loading cannot be abstracted as it depends on the reader and the rules.
        pass



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
import datetime
import re
from pathlib import Path
from typing import Dict, List, Callable, Optional, Any, Union

from pyshacl.consts import SH
from rdflib import URIRef, Literal
from rdflib.namespace import XSD

from kgforge.core import Resource
from kgforge.core.archetypes import Model, Store
from kgforge.core.commons.actions import Action
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import ValidationError
from kgforge.core.commons.execution import run
from kgforge.specializations.models.rdf.collectors import NodeProperties
from kgforge.specializations.models.rdf.directory_service import DirectoryService
from kgforge.specializations.models.rdf.service import RdfService
from kgforge.specializations.models.rdf.store_service import StoreService
from kgforge.specializations.models.rdf.utils import as_term

DEFAULT_VALUE = {
    XSD.string: str(),
    XSD.normalizedString: str(),
    XSD.anyURI: str(),
    XSD.float: float(),
    XSD.double: float(),
    XSD.decimal: int(),
    XSD.int: int(),
    XSD.integer: int(),
    XSD.positiveInteger: int(),
    XSD.negativeInteger: int(),
    XSD.nonPositiveInteger: int(),
    XSD.nonNegativeInteger: int(),
    XSD.long: int(),
    XSD.short: int(),
    XSD.unsignedLong: int(),
    XSD.unsignedInt: int(),
    XSD.unsignedShort: int(),
    XSD.byte: int(),
    XSD.unsignedByte: int(),
    XSD.base64Binary: int(),
    XSD.boolean: bool(),
    XSD.time: datetime.time().isoformat(),
    XSD.date: datetime.date(9999, 12, 31).isoformat(),
    XSD.dateTime: datetime.datetime(9999, 12, 31).isoformat(),
}

DEFAULT_TYPE_ORDER = [str, float, int, bool, datetime.date, datetime.time]


class RdfModel(Model):
    """Specialization of Model that follows SHACL shapes"""

    def __init__(self, source: str, **source_config) -> None:
        super().__init__(source, **source_config)

    # Vocabulary.

    def _prefixes(self) -> Dict[str, str]:
        return self.service.context.prefixes

    def _types(self) -> List[str]:
        return list(self.service.types_to_shapes.keys())

    def context(self) -> Context:
        return self.service.context

    def resolve_context(self, iri: str) -> Dict:
        return self.service.resolve_context(iri)

    def _generate_context(self) -> Context:
        document = self.service.generate_context()
        if document:
            return Context(document)

    # Templates.

    def _template(self, type: str, only_required: bool) -> Dict:
        try:
            uri = self.service.types_to_shapes[type]
        except KeyError:
            raise ValueError("type not found")
        node_properties = self.service.materialize(uri)
        dictionary = parse_attributes(node_properties, only_required, None)
        return dictionary

    # Validation.

    def schema_id(self, type: str) -> str:
        try:
            shape_iri = self.service.types_to_shapes[type]
            return str(self.service.schema_source_id(shape_iri))
        except KeyError:
            raise ValueError("type not found")

    def validate(self, data: Union[Resource, List[Resource]], execute_actions_before: bool) -> None:
        run(self._validate_one, self._validate_many, data, execute_actions=execute_actions_before,
            exception=ValidationError, monitored_status="_validated")

    def _validate_many(self, resources: List[Resource]) -> None:
        for resource in resources:
            conforms, graph, _ = self.service.validate(resource)
            if conforms:
                resource._validated = True
                action = Action(self._validate_many.__name__, conforms, None)
            else:
                resource._validated = False
                violations = set(" ".join(re.findall('[A-Z][^A-Z]*', as_term(o)))
                                 for o in graph.objects(None, SH.sourceConstraintComponent))
                message = f"violation(s) of type(s) {', '.join(sorted(violations))}"
                action = Action(self._validate_many.__name__, conforms, ValidationError(message))
            resource._last_action = action

    def _validate_one(self, resource: Resource) -> None:
        conforms, _, report = self.service.validate(resource)
        if conforms is False:
            raise ValidationError("\n" + report)

    # Utils.

    @staticmethod
    def _service_from_directory(dirpath: Path, context_iri: str, **dir_config) -> RdfService:
        return DirectoryService(dirpath, context_iri)

    @staticmethod
    def _service_from_store(store: Callable, context_config: Optional[Dict], **source_config) -> Any:
        endpoint = source_config.get("endpoint")
        token = source_config.get("token")
        bucket = source_config["bucket"]
        default_store: Store = store(endpoint, bucket, token)

        if context_config:
            context_endpoint = context_config.get("endpoint", default_store.endpoint)
            context_token = context_config.get("token", default_store.token)
            context_bucket = context_config.get("bucket", default_store.bucket)
            context_iri = context_config.get("iri")
            if (context_endpoint != default_store.endpoint
                    or context_token != default_store.token
                    or context_bucket != default_store.bucket):
                context_store: Store = store(context_endpoint, context_bucket, context_token)
                service = StoreService(default_store, context_iri, context_store)
            else:
                service = StoreService(default_store, context_iri, None)
        else:
            service = StoreService(default_store)

        return service


def parse_attributes(node: NodeProperties, only_required: bool,
                     inherited_constraint: Optional[str]) -> Dict:
    attributes = dict()
    if hasattr(node, "path"):
        if only_required is True:
            if not hasattr(node, "mandatory"):
                return attributes
        if hasattr(node, "properties"):
            parent_constraint = node.constraint if hasattr(node, "constraint") else None
            v = parse_properties(node.properties, only_required, parent_constraint)
        else:
            v = parse_value(node, inherited_constraint)
        attributes[as_term(node.path)] = v
    elif hasattr(node, "properties"):
        parent_constraint = node.constraint if hasattr(node, "constraint") else None
        attributes.update(parse_properties(node.properties, only_required, parent_constraint))
    return attributes


def parse_properties(items: List[NodeProperties], only_required: bool, inherited_constraint: str) -> Dict:
    props = dict()
    for item in items:
        props.update(parse_attributes(item, only_required, inherited_constraint))
    return props


def parse_value(node: NodeProperties, parent_constraint: str) -> Any:
    v = ""
    if hasattr(node, "values"):
        # node.constraint could be: in, or, xor
        if hasattr(node, "constraint") or parent_constraint:
            constraint = parent_constraint or node.constraint
            if constraint in ("in", "or", "xone"):
                v = default_values(node.values, one=False)
            else:
                v = default_values(node.values, one=True)
        else:
            v = default_values(node.values, one=True)
    return v


def default_values(values, one: bool):
    if isinstance(values, list):
        all_default_values = [default_value(v) for v in values]
        if all_default_values:
            if one:
                for data_type in DEFAULT_TYPE_ORDER:
                    for val in all_default_values:
                        if isinstance(val, data_type):
                            return val
            else:
                first_type = type(all_default_values[0])
                sortable = all(isinstance(v, first_type) for v in all_default_values)
                if sortable:
                    return sorted(all_default_values)
                else:
                    types_position = {DEFAULT_TYPE_ORDER.index(type(v)): v for v in all_default_values}
                    return [types_position[k] for k in sorted(types_position.keys())]
    else:
        return default_value(values)


def default_value(value):
    # TODO: replace the as_term function with context.to_symbol
    if value in DEFAULT_VALUE:
        return DEFAULT_VALUE[value]
    elif isinstance(value, URIRef):
        return as_term(value)
    elif isinstance(value, Literal):
        return value.toPython()
    else:
        return value


def object_value(value):
    return {"type": as_term(value)}


def data_value(value):
    if value in DEFAULT_VALUE:
        return DEFAULT_VALUE[value]
    else:
        return value
